"""
pyrad.proc.process_phase
======================================

Functions for PhiDP and KDP processing and attenuation correction

.. autosummary::
    :toctree: generated/

    process_correct_phidp0
    process_smooth_phidp_single_window
    process_smooth_phidp_double_window
    process_kdp_leastsquare_single_window
    process_kdp_leastsquare_double_window
    process_phidp_kdp_Vulpiani
    process_phidp_kdp_Kalman
    process_phidp_kdp_Maesaka
    process_phidp_kdp_lp
    process_selfconsistency_kdp_phidp
    process_selfconsistency_bias
    process_attenuation

"""

from copy import deepcopy
from warnings import warn

import numpy as np

import pyart

from ..io.io_aux import get_datatype_fields


def process_correct_phidp0(procstatus, dscfg, radar_list=None):
    """
    corrects phidp of the system phase

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types
        rmin : float. Dataset keyword
            The minimum range where to look for valid data [m]
        rmax : float. Dataset keyword
            The maximum range where to look for valid data [m]
        rcell : float. Dataset keyword
            The length of a continuous cell to consider it valid precip [m]
        Zmin : float. Dataset keyword
            The minimum reflectivity [dBZ]
        Zmax : float. Dataset keyword
            The maximum reflectivity [dBZ]
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'dBZ':
            refl_field = 'reflectivity'
        if datatype == 'dBZc':
            refl_field = 'corrected_reflectivity'
        if datatype == 'PhiDP':
            psidp_field = 'differential_phase'
        if datatype == 'PhiDPc':
            psidp_field = 'corrected_differential_phase'
        if datatype == 'uPhiDP':
            psidp_field = 'uncorrected_differential_phase'

    ind_rad = int(radarnr[5:8])-1
    if radar_list[ind_rad] is None:
        warn('No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if (refl_field not in radar.fields) or (psidp_field not in radar.fields):
        warn('Unable to correct PhiDP system offset. Missing data')
        return None, None

    ind_rmin = np.where(radar.range['data'] > dscfg['rmin'])[0][0]
    ind_rmax = np.where(radar.range['data'] < dscfg['rmax'])[0][-1]
    r_res = radar.range['data'][1]-radar.range['data'][0]
    min_rcons = int(dscfg['rcell']/r_res)

    if psidp_field.startswith('corrected_'):
        phidp_field = psidp_field
    elif psidp_field.startswith('uncorrected_'):
        phidp_field = psidp_field.replace('uncorrected_', 'corrected_', 1)
    else:
        phidp_field = 'corrected_'+psidp_field

    phidp = pyart.correct.correct_sys_phase(
        radar, ind_rmin=ind_rmin, ind_rmax=ind_rmax, min_rcons=min_rcons,
        zmin=dscfg['Zmin'], zmax=dscfg['Zmax'], psidp_field=psidp_field,
        refl_field=refl_field, phidp_field=phidp_field)

    # prepare for exit
    new_dataset = {'radar_out': deepcopy(radar)}
    new_dataset['radar_out'].fields = dict()
    new_dataset['radar_out'].add_field(phidp_field, phidp)

    return new_dataset, ind_rad


def process_smooth_phidp_single_window(procstatus, dscfg, radar_list=None):
    """
    corrects phidp of the system phase and smoothes it using one window

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types
        rmin : float. Dataset keyword
            The minimum range where to look for valid data [m]
        rmax : float. Dataset keyword
            The maximum range where to look for valid data [m]
        rcell : float. Dataset keyword
            The length of a continuous cell to consider it valid precip [m]
        rwind : float. Dataset keyword
            The length of the smoothing window [m]
        Zmin : float. Dataset keyword
            The minimum reflectivity [dBZ]
        Zmax : float. Dataset keyword
            The maximum reflectivity [dBZ]
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'dBZ':
            refl_field = 'reflectivity'
        if datatype == 'dBZc':
            refl_field = 'corrected_reflectivity'
        if datatype == 'PhiDP':
            psidp_field = 'differential_phase'
        if datatype == 'PhiDPc':
            psidp_field = 'corrected_differential_phase'
        if datatype == 'uPhiDP':
            psidp_field = 'uncorrected_differential_phase'

    ind_rad = int(radarnr[5:8])-1
    if radar_list[ind_rad] is None:
        warn('No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if (refl_field not in radar.fields) or (psidp_field not in radar.fields):
        warn('Unable to smooth PhiDP. Missing data')
        return None, None

    ind_rmin = np.where(radar.range['data'] > dscfg['rmin'])[0][0]
    ind_rmax = np.where(radar.range['data'] < dscfg['rmax'])[0][-1]
    r_res = radar.range['data'][1]-radar.range['data'][0]
    min_rcons = int(dscfg['rcell']/r_res)
    wind_len = int(dscfg['rwind']/r_res)
    min_valid = int(wind_len/2+1)

    if psidp_field.startswith('corrected_'):
        phidp_field = psidp_field
    elif psidp_field.startswith('uncorrected_'):
        phidp_field = psidp_field.replace('uncorrected_', 'corrected_', 1)
    else:
        phidp_field = 'corrected_'+psidp_field

    phidp = pyart.correct.smooth_phidp_single_window(
        radar, ind_rmin=ind_rmin, ind_rmax=ind_rmax, min_rcons=min_rcons,
        zmin=dscfg['Zmin'], zmax=dscfg['Zmax'], wind_len=wind_len,
        min_valid=min_valid, psidp_field=psidp_field, refl_field=refl_field,
        phidp_field=phidp_field)

    # prepare for exit
    new_dataset = {'radar_out': deepcopy(radar)}
    new_dataset['radar_out'].fields = dict()
    new_dataset['radar_out'].add_field(phidp_field, phidp)

    return new_dataset, ind_rad


def process_smooth_phidp_double_window(procstatus, dscfg, radar_list=None):
    """
    corrects phidp of the system phase and smoothes it using one window

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types
        rmin : float. Dataset keyword
            The minimum range where to look for valid data [m]
        rmax : float. Dataset keyword
            The maximum range where to look for valid data [m]
        rcell : float. Dataset keyword
            The length of a continuous cell to consider it valid precip [m]
        rwinds : float. Dataset keyword
            The length of the short smoothing window [m]
        rwindl : float. Dataset keyword
            The length of the long smoothing window [m]
        Zmin : float. Dataset keyword
            The minimum reflectivity [dBZ]
        Zmax : float. Dataset keyword
            The maximum reflectivity [dBZ]
        Zthr : float. Dataset keyword
            The threshold defining wich smoothed data to used [dBZ]
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'dBZ':
            refl_field = 'reflectivity'
        if datatype == 'dBZc':
            refl_field = 'corrected_reflectivity'
        if datatype == 'PhiDP':
            psidp_field = 'differential_phase'
        if datatype == 'PhiDPc':
            psidp_field = 'corrected_differential_phase'
        if datatype == 'uPhiDP':
            psidp_field = 'uncorrected_differential_phase'

    ind_rad = int(radarnr[5:8])-1
    if radar_list[ind_rad] is None:
        warn('No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if (refl_field not in radar.fields) or (psidp_field not in radar.fields):
        warn('Unable to smooth PhiDP. Missing data')
        return None, None

    ind_rmin = np.where(radar.range['data'] > dscfg['rmin'])[0][0]
    ind_rmax = np.where(radar.range['data'] < dscfg['rmax'])[0][-1]
    r_res = radar.range['data'][1]-radar.range['data'][0]
    min_rcons = int(dscfg['rcell']/r_res)
    swind_len = int(dscfg['rwinds']/r_res)
    smin_valid = int(swind_len/2+1)
    lwind_len = int(dscfg['rwindl']/r_res)
    lmin_valid = int(lwind_len/2+1)

    if psidp_field.startswith('corrected_'):
        phidp_field = psidp_field
    elif psidp_field.startswith('uncorrected_'):
        phidp_field = psidp_field.replace('uncorrected_', 'corrected_', 1)
    else:
        phidp_field = 'corrected_'+psidp_field

    phidp = pyart.correct.smooth_phidp_double_window(
        radar, ind_rmin=ind_rmin, ind_rmax=ind_rmax, min_rcons=min_rcons,
        zmin=dscfg['Zmin'], zmax=dscfg['Zmax'], swind_len=swind_len,
        smin_valid=smin_valid, lwind_len=lwind_len, lmin_valid=lmin_valid,
        zthr=dscfg['Zthr'], psidp_field=psidp_field, refl_field=refl_field,
        phidp_field=phidp_field)

    # prepare for exit
    new_dataset = {'radar_out': deepcopy(radar)}
    new_dataset['radar_out'].fields = dict()
    new_dataset['radar_out'].add_field(phidp_field, phidp)

    return new_dataset, ind_rad


def process_phidp_kdp_Maesaka(procstatus, dscfg, radar_list=None):
    """
    Estimates PhiDP and KDP using the method by Maesaka. This method only
    retrieves data in rain (i.e. below the melting layer)

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types
        rmin : float. Dataset keyword
            The minimum range where to look for valid data [m]
        rmax : float. Dataset keyword
            The maximum range where to look for valid data [m]
        rcell : float. Dataset keyword
            The length of a continuous cell to consider it valid precip [m]
        Zmin : float. Dataset keyword
            The minimum reflectivity [dBZ]
        Zmax : float. Dataset keyword
            The maximum reflectivity [dBZ]
        fzl : float. Dataset keyword
            The freezing level height [m]. Default 2000.
        ml_thickness : float. Dataset keyword
            The melting layer thickness in meters. Default 700.
        beamwidth : float. Dataset keyword
            the antenna beamwidth [deg]. If None that of the keys
            radar_beam_width_h or radar_beam_width_v in attribute
            instrument_parameters of the radar object will be used. If the key
            or the attribute are not present the beamwidth will be set to None
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    temp_field = None
    iso0_field = None
    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'PhiDP':
            psidp_field = 'differential_phase'
        if datatype == 'PhiDPc':
            psidp_field = 'corrected_differential_phase'
        if datatype == 'uPhiDP':
            psidp_field = 'uncorrected_differential_phase'
        if datatype == 'dBZ':
            refl_field = 'reflectivity'
        if datatype == 'dBZc':
            refl_field = 'corrected_reflectivity'
        if datatype == 'TEMP':
            temp_field = 'temperature'
        if datatype == 'H_ISO0':
            iso0_field = 'height_over_iso0'

    ind_rad = int(radarnr[5:8])-1
    if radar_list[ind_rad] is None:
        warn('No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if ((refl_field not in radar.fields) or
            (psidp_field not in radar.fields)):
        warn('Unable to retrieve PhiDP KDP using the Maesaka approach. ' +
             'Missing data')
        return None, None

    # determine which freezing level reference
    temp_ref = 'temperature'
    if temp_field is None and iso0_field is None:
        warn('Field to obtain the freezing level was not specified. ' +
             'Using fixed freezing level height')
        temp_ref = 'fixed_fzl'
    elif temp_field is not None:
        if temp_field not in radar.fields:
            warn('COSMO temperature field not available. ' +
                 'Using fixed freezing level height')
            temp_ref = 'fixed_fzl'
    elif iso0_field is not None:
        if iso0_field not in radar.fields:
            warn('Height over iso0 field not available. ' +
                 'Using fixed freezing level height')
            temp_ref = 'fixed_fzl'
        else:
            temp_ref = 'height_over_iso0'

    # determine freezing level height if necessary
    fzl = None
    if temp_ref == 'fixed_fzl':
        if 'fzl' in dscfg:
            fzl = dscfg['fzl']
        else:
            fzl = 2000.
            warn('Freezing level height not defined. Using default ' +
                 str(fzl)+' m')

    thickness = 700.
    if 'ml_thickness' in dscfg:
        thickness = dscfg['ml_thickness']

    phidp_field = 'corrected_differential_phase'
    kdp_field = 'corrected_specific_differential_phase'

    radar_aux = deepcopy(radar)

    # correct PhiDP0
    ind_rmin = np.where(radar_aux.range['data'] > dscfg['rmin'])[0][0]
    ind_rmax = np.where(radar_aux.range['data'] < dscfg['rmax'])[0][-1]
    r_res = radar_aux.range['data'][1]-radar_aux.range['data'][0]
    min_rcons = int(dscfg['rcell']/r_res)

    phidp = pyart.correct.correct_sys_phase(
        radar_aux, ind_rmin=ind_rmin, ind_rmax=ind_rmax, min_rcons=min_rcons,
        zmin=dscfg['Zmin'], zmax=dscfg['Zmax'], psidp_field=psidp_field,
        refl_field=refl_field, phidp_field=phidp_field)

    # filter out data in an above the melting layer
    mask = np.ma.getmaskarray(phidp['data'])

    beamwidth = dscfg.get('beamwidth', None)
    if beamwidth is None:
        if radar.instrument_parameters is not None:
            if 'radar_beam_width_h' in radar.instrument_parameters:
                beamwidth = radar.instrument_parameters['radar_beam_width_h'][
                    'data'][0]
            elif 'radar_beam_width_v' in radar.instrument_parameters:
                beamwidth = radar.instrument_parameters['radar_beam_width_v'][
                    'data'][0]
    if beamwidth is None:
        warn('Antenna beam width unknown.')

    mask_fzl, _ = pyart.correct.get_mask_fzl(
        radar_aux, fzl=fzl, doc=15, min_temp=0., max_h_iso0=0.,
        thickness=thickness, beamwidth=beamwidth, temp_field=temp_field,
        iso0_field=iso0_field, temp_ref=temp_ref)
    mask = np.logical_or(mask, mask_fzl)

    # filter out data with invalid reflectivity
    mask_refl = np.ma.getmaskarray(radar_aux.fields[refl_field]['data'])
    mask = np.logical_or(mask, mask_refl)

    phidp['data'] = np.ma.masked_where(mask, phidp['data'])
    fill_value = pyart.config.get_fillvalue()
    radar_aux.add_field(phidp_field, phidp, replace_existing=True)

    # the return data is not a masked array
    kdp, phidpf, _ = pyart.retrieve.kdp_proc.kdp_maesaka(
        radar_aux, gatefilter=None, method='cg', backscatter=None, Clpf=1.,
        length_scale=None, first_guess=0.01, finite_order='low',
        fill_value=fill_value, psidp_field=phidp_field, kdp_field=kdp_field,
        phidp_field=phidp_field)

    kdp['data'] = np.ma.masked_where(mask, kdp['data'])
    phidpf['data'] = np.ma.masked_where(mask, phidpf['data'])

    # prepare for exit
    new_dataset = {'radar_out': deepcopy(radar_aux)}
    new_dataset['radar_out'].fields = dict()
    new_dataset['radar_out'].add_field(phidp_field, phidpf)
    new_dataset['radar_out'].add_field(kdp_field, kdp)

    return new_dataset, ind_rad


def process_phidp_kdp_lp(procstatus, dscfg, radar_list=None):
    """
    Estimates PhiDP and KDP using a linear programming algorithm.
    This method only retrieves data in rain (i.e. below the melting layer)

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types
        fzl : float. Dataset keyword
            The freezing level height [m]. Default 2000.
        ml_thickness : float. Dataset keyword
            The melting layer thickness in meters. Default 700.
        beamwidth : float. Dataset keyword
            the antenna beamwidth [deg]. If None that of the keys
            radar_beam_width_h or radar_beam_width_v in attribute
            instrument_parameters of the radar object will be used. If the key
            or the attribute are not present the beamwidth will be set to None
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    temp_field = None
    iso0_field = None
    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'PhiDP':
            psidp_field = 'differential_phase'
        if datatype == 'PhiDPc':
            psidp_field = 'corrected_differential_phase'
        if datatype == 'uPhiDP':
            psidp_field = 'uncorrected_differential_phase'
        if datatype == 'dBZ':
            refl_field = 'reflectivity'
        if datatype == 'dBZc':
            refl_field = 'corrected_reflectivity'
        if datatype == 'RhoHVc':
            rhv_field = 'corrected_cross_correlation_ratio'
        if datatype == 'RhoHV':
            rhv_field = 'cross_correlation_ratio'
        if datatype == 'SNRh':
            snr_field = 'signal_to_noise_ratio_hh'
        if datatype == 'TEMP':
            temp_field = 'temperature'
        if datatype == 'H_ISO0':
            iso0_field = 'height_over_iso0'

    ind_rad = int(radarnr[5:8])-1
    if radar_list[ind_rad] is None:
        warn('No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if ((refl_field not in radar.fields) or
            (psidp_field not in radar.fields) or
            (rhv_field not in radar.fields) or
            (snr_field not in radar.fields)):
        warn('Unable to retrieve PhiDP and KDP using the LP approach. ' +
             'Missing data')
        return None, None

    # determine which freezing level reference
    temp_ref = 'temperature'
    if temp_field is None and iso0_field is None:
        warn('Field to obtain the freezing level was not specified. ' +
             'Using fixed freezing level height')
        temp_ref = 'fixed_fzl'
    elif temp_field is not None:
        if temp_field not in radar.fields:
            warn('COSMO temperature field not available. ' +
                 'Using fixed freezing level height')
            temp_ref = 'fixed_fzl'
    elif iso0_field is not None:
        if iso0_field not in radar.fields:
            warn('Height over iso0 field not available. ' +
                 'Using fixed freezing level height')
            temp_ref = 'fixed_fzl'
        else:
            temp_ref = 'height_over_iso0'

    # determine freezing level height if necessary
    fzl = None
    if temp_ref == 'fixed_fzl':
        if 'fzl' in dscfg:
            fzl = dscfg['fzl']
        else:
            fzl = 2000.
            warn('Freezing level height not defined. Using default ' +
                 str(fzl)+' m')

    thickness = 700.
    if 'ml_thickness' in dscfg:
        thickness = dscfg['ml_thickness']

    radar_aux = deepcopy(radar)

    # user config
    LP_solver = dscfg.get('LP_solver', 'cvxopt')

    # filter out data in an above the melting layer
    mask = np.ma.getmaskarray(radar_aux.fields[psidp_field]['data'])
    beamwidth = dscfg.get('beamwidth', None)
    if beamwidth is None:
        if radar.instrument_parameters is not None:
            if 'radar_beam_width_h' in radar.instrument_parameters:
                beamwidth = radar.instrument_parameters[
                    'radar_beam_width_h']['data'][0]
            elif 'radar_beam_width_v' in radar.instrument_parameters:
                beamwidth = radar.instrument_parameters[
                    'radar_beam_width_v']['data'][0]
    if beamwidth is None:
        warn('Antenna beam width unknown.')

    mask_fzl, _ = pyart.correct.get_mask_fzl(
        radar_aux, fzl=fzl, doc=15, min_temp=0., max_h_iso0=0.,
        thickness=thickness, beamwidth=beamwidth, temp_field=temp_field,
        iso0_field=iso0_field, temp_ref=temp_ref)
    mask = np.logical_or(mask, mask_fzl)

    # filter out data with invalid reflectivity
    mask_refl = np.ma.getmaskarray(radar_aux.fields[refl_field]['data'])
    mask = np.logical_or(mask, mask_refl)

    radar_aux.fields[psidp_field]['data'] = np.ma.masked_where(
        mask, radar_aux.fields[psidp_field]['data'])

    phidp_field = 'corrected_differential_phase'
    kdp_field = 'corrected_specific_differential_phase'

    phidp, kdp = pyart.correct.phase_proc_lp(
        radar_aux, 0, debug=False, self_const=60000.0,
        low_z=10.0, high_z=53.0, min_phidp=0.01, min_ncp=10.,
        min_rhv=0.6, fzl=4000.0, sys_phase=0.0,
        overide_sys_phase=True, nowrap=None, really_verbose=False,
        LP_solver=LP_solver, refl_field=refl_field, ncp_field=snr_field,
        rhv_field=rhv_field, phidp_field=psidp_field, kdp_field=kdp_field,
        unf_field=phidp_field, window_len=35, proc=1)

    kdp['data'] = np.ma.masked_where(mask, kdp['data'])
    phidp['data'] = np.ma.masked_where(mask, phidp['data'])

    # prepare for exit
    new_dataset = {'radar_out': deepcopy(radar)}
    new_dataset['radar_out'].fields = dict()
    new_dataset['radar_out'].add_field(phidp_field, phidp)
    new_dataset['radar_out'].add_field(kdp_field, kdp)

    return new_dataset, ind_rad


def process_kdp_leastsquare_single_window(procstatus, dscfg, radar_list=None):
    """
    Computes specific differential phase using a piecewise least square method

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types
        rwind : float. Dataset keyword
            The length of the segment for the least square method [m]
        vectorize : bool. Dataset keyword
            Whether to vectorize the KDP processing. Default false
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'PhiDP':
            phidp_field = 'differential_phase'
        if datatype == 'PhiDPc':
            phidp_field = 'corrected_differential_phase'
        if datatype == 'uPhiDP':
            phidp_field = 'uncorrected_differential_phase'

    ind_rad = int(radarnr[5:8])-1
    if radar_list[ind_rad] is None:
        warn('No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if phidp_field not in radar.fields:
        warn('Unable to retrieve KDP from PhiDP using least square. ' +
             'Missing data')
        return None, None

    r_res = radar.range['data'][1]-radar.range['data'][0]
    wind_len = int(dscfg['rwind']/r_res)
    min_valid = int(wind_len/2+1)
    kdp_field = 'corrected_specific_differential_phase'
    vectorize = dscfg.get('vectorize', False)

    kdp = pyart.retrieve.kdp_leastsquare_single_window(
        radar, wind_len=wind_len, min_valid=min_valid, phidp_field=phidp_field,
        kdp_field=kdp_field, vectorize=vectorize)

    # prepare for exit
    new_dataset = {'radar_out': deepcopy(radar)}
    new_dataset['radar_out'].fields = dict()
    new_dataset['radar_out'].add_field(kdp_field, kdp)

    return new_dataset, ind_rad


def process_kdp_leastsquare_double_window(procstatus, dscfg, radar_list=None):
    """
    Computes specific differential phase using a piecewise least square method

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types
        rwinds : float. Dataset keyword
            The length of the short segment for the least square method [m]
        rwindl : float. Dataset keyword
            The length of the long segment for the least square method [m]
        Zthr : float. Dataset keyword
            The threshold defining which estimated data to use [dBZ]
        vectorize : Bool. Dataset keyword
            Whether to vectorize the KDP processing. Default false
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'PhiDP':
            phidp_field = 'differential_phase'
        if datatype == 'PhiDPc':
            phidp_field = 'corrected_differential_phase'
        if datatype == 'uPhiDP':
            phidp_field = 'uncorrected_differential_phase'
        if datatype == 'dBZ':
            refl_field = 'reflectivity'
        if datatype == 'dBZc':
            refl_field = 'corrected_reflectivity'

    ind_rad = int(radarnr[5:8])-1
    if radar_list[ind_rad] is None:
        warn('No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if (phidp_field not in radar.fields) or (refl_field not in radar.fields):
        warn('Unable to retrieve KDP from PhiDP using least square. ' +
             'Missing data')
        return None, None

    r_res = radar.range['data'][1]-radar.range['data'][0]
    swind_len = int(dscfg['rwinds']/r_res)
    smin_valid = int(swind_len/2+1)
    lwind_len = int(dscfg['rwindl']/r_res)
    lmin_valid = int(lwind_len/2+1)
    vectorize = dscfg.get('vectorize', False)

    kdp_field = 'corrected_specific_differential_phase'

    kdp = pyart.retrieve.kdp_leastsquare_double_window(
        radar, swind_len=swind_len, smin_valid=smin_valid,
        lwind_len=lwind_len, lmin_valid=lmin_valid, zthr=dscfg['Zthr'],
        phidp_field=phidp_field, refl_field=refl_field, kdp_field=kdp_field,
        vectorize=vectorize)

    # prepare for exit
    new_dataset = {'radar_out': deepcopy(radar)}
    new_dataset['radar_out'].fields = dict()
    new_dataset['radar_out'].add_field(kdp_field, kdp)

    return new_dataset, ind_rad


def process_phidp_kdp_Vulpiani(procstatus, dscfg, radar_list=None):
    """
    Computes specific differential phase and differential phase using the
    method developed by Vulpiani et al.
    The data is assumed to be clutter free and monotonous

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types
        rwind : float. Dataset keyword
            The length of the segment [m]
        n_iter : int. Dataset keyword
            number of iterations
        interp : boolean. Dataset keyword
            if set non valid values are interpolated using neighbouring valid
            values
        parallel : boolean. Dataset keyword
            if set use parallel computing
        get_phidp : boolean. Datset keyword
            if set the PhiDP computed by integrating the resultant KDP is
            added to the radar field
        frequency : float. Dataset keyword
            the radar frequency [Hz]. If None that of the key
            frequency in attribute instrument_parameters of the radar
            object will be used. If the key or the attribute are not present
            it will be assumed that the radar is C band
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'PhiDP':
            phidp_field = 'differential_phase'
        if datatype == 'PhiDPc':
            phidp_field = 'corrected_differential_phase'
        if datatype == 'uPhiDP':
            phidp_field = 'uncorrected_differential_phase'

    ind_rad = int(radarnr[5:8])-1
    if radar_list[ind_rad] is None:
        warn('No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if phidp_field not in radar.fields:
        warn('Unable to retrieve KDP from PhiDP using least square. ' +
             'Missing data')
        return None, None

    # number of iterations
    n_iter = 3
    if 'n_iter' in dscfg:
        n_iter = dscfg['n_iter']

    # window length (must be even)
    r_res = radar.range['data'][1]-radar.range['data'][0]
    wind_len = int(dscfg['rwind']/r_res)
    if wind_len % 2 == 1:
        wind_len += 1

    # interpolate invalid values?
    interp = 0
    if 'interp' in dscfg:
        interp = dscfg['interp']

    # parallel computing?
    parallel = 1
    if 'parallel' in dscfg:
        parallel = dscfg['parallel']

    # get PhiDP computed from KDP?
    get_phidp = 0
    if 'get_phidp' in dscfg:
        get_phidp = dscfg['get_phidp']

    # get band
    freq = dscfg.get('frequency', None)
    if freq is None:
        if (radar.instrument_parameters is not None and
                'frequency' in radar.instrument_parameters):
            freq = radar.instrument_parameters['frequency']['data'][0]

    band = 'C'
    if freq is None:
        warn('Radar frequency unknown. Default C band will be applied')
    else:
        band = pyart.retrieve.get_freq_band(freq)
        if band is None:
            if freq < 2e9:
                band = 'S'
            elif freq > 12e9:
                band = 'X'
            warn('Radar frequency out of range. Valid bands are S, C or X. ' +
                 band + ' band will be applied')

    kdp_field = 'corrected_specific_differential_phase'
    phidpr_field = 'corrected_differential_phase'

    kdp_dict, phidpr_dict = pyart.retrieve.kdp_vulpiani(
        radar, gatefilter=None, fill_value=None, psidp_field=phidp_field,
        kdp_field=kdp_field, phidp_field=phidpr_field, band=band,
        windsize=wind_len, n_iter=n_iter, interp=interp,
        prefilter_psidp=False, filter_opt=None, parallel=parallel)

    # prepare for exit
    new_dataset = {'radar_out': deepcopy(radar)}
    new_dataset['radar_out'].fields = dict()
    new_dataset['radar_out'].add_field(kdp_field, kdp_dict)
    if get_phidp:
        new_dataset['radar_out'].add_field(phidpr_field, phidpr_dict)

    return new_dataset, ind_rad


def process_phidp_kdp_Kalman(procstatus, dscfg, radar_list=None):
    """
    Computes specific differential phase and differential phase using the
    Kalman filter as proposed by Schneebeli et al.
    The data is assumed to be clutter free and continous

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types
        parallel : boolean. Dataset keyword
            if set use parallel computing
        get_phidp : boolean. Datset keyword
            if set the PhiDP computed by integrating the resultant KDP is
            added to the radar field
        frequency : float. Dataset keyword
            the radar frequency [Hz]. If None that of the key
            frequency in attribute instrument_parameters of the radar
            object will be used. If the key or the attribute are not present
            it will be assumed that the radar is C band
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'PhiDP':
            phidp_field = 'differential_phase'
        if datatype == 'PhiDPc':
            phidp_field = 'corrected_differential_phase'
        if datatype == 'uPhiDP':
            phidp_field = 'uncorrected_differential_phase'

    ind_rad = int(radarnr[5:8])-1
    if radar_list[ind_rad] is None:
        warn('No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if phidp_field not in radar.fields:
        warn('Unable to retrieve KDP from PhiDP using least square. ' +
             'Missing data')
        return None, None

    # User defined options
    parallel = dscfg.get('parallel', 1)
    get_phidp = dscfg.get('get_phidp', 0)

    # get band
    freq = dscfg.get('frequency', None)
    if freq is None:
        if (radar.instrument_parameters is not None and
                'frequency' in radar.instrument_parameters):
            freq = radar.instrument_parameters['frequency']['data'][0]

    band = 'C'
    if freq is None:
        warn('Radar frequency unknown. Default C band will be applied')
    else:
        band = pyart.retrieve.get_freq_band(freq)
        if band is None:
            if freq < 2e9:
                band = 'S'
            elif freq > 12e9:
                band = 'X'
            warn('Radar frequency out of range. Valid bands are S, C or X. ' +
                 band + ' band will be applied')

    kdp_field = 'corrected_specific_differential_phase'
    phidpr_field = 'corrected_differential_phase'

    kdp_dict, _, phidpr_dict = pyart.retrieve.kdp_schneebeli(
        radar, gatefilter=None, fill_value=None, psidp_field=phidp_field,
        kdp_field=kdp_field, phidp_field=phidpr_field, band=band, rcov=0,
        pcov=0, prefilter_psidp=False, filter_opt=None, parallel=parallel)

    # prepare for exit
    new_dataset = {'radar_out': deepcopy(radar)}
    new_dataset['radar_out'].fields = dict()
    new_dataset['radar_out'].add_field(kdp_field, kdp_dict)
    if get_phidp:
        new_dataset['radar_out'].add_field(phidpr_field, phidpr_dict)

    return new_dataset, ind_rad


def process_attenuation(procstatus, dscfg, radar_list=None):
    """
    Computes specific attenuation and specific differential attenuation using
    the Z-Phi method and corrects reflectivity and differential reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted Configuration Keywords::

        datatype : list of string. Dataset keyword
            The input data types
        ATT_METHOD : float. Dataset keyword
            The attenuation estimation method used. One of the following:
            ZPhi, Philin
        fzl : float. Dataset keyword
            The default freezing level height. It will be used if no
            temperature field name is specified or the temperature field is
            not in the radar object. Default 2000.
    radar_list : list of Radar objects
        Optional. list of radar objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    temp = None
    iso0 = None
    zdr = None
    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'dBZc':
            refl = 'corrected_reflectivity'
        if datatype == 'PhiDPc':
            phidp = 'corrected_differential_phase'
        if datatype == 'ZDRc':
            zdr = 'corrected_differential_reflectivity'
        if datatype == 'dBZ':
            refl = 'reflectivity'
        if datatype == 'PhiDP':
            phidp = 'differential_phase'
        if datatype == 'ZDR':
            zdr = 'differential_reflectivity'
        if datatype == 'TEMP':
            temp = 'temperature'
        if datatype == 'H_ISO0':
            iso0 = 'height_over_iso0'

    ind_rad = int(radarnr[5:8])-1
    if radar_list[ind_rad] is None:
        warn('No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if phidp not in radar.fields or refl not in radar.fields:
        warn('Unable to compute attenuation. Missing data')
        return None, None

    # determine which freezing level reference
    temp_ref = 'temperature'
    if temp is None and iso0 is None:
        warn('Field to obtain the freezing level was not specified. ' +
             'Using fixed freezing level height')
        temp_ref = 'fixed_fzl'
    elif temp is not None:
        if temp not in radar.fields:
            warn('COSMO temperature field not available. ' +
                 'Using fixed freezing level height')
            temp_ref = 'fixed_fzl'
    elif iso0 is not None:
        if iso0 not in radar.fields:
            warn('Height over iso0 field not available. ' +
                 'Using fixed freezing level height')
            temp_ref = 'fixed_fzl'
        else:
            temp_ref = 'height_over_iso0'

    # determine freezing level height if necessary
    fzl = None
    if temp_ref == 'fixed_fzl':
        if 'fzl' in dscfg:
            fzl = dscfg['fzl']
        else:
            fzl = 2000.
            warn('Freezing level height not defined. Using default ' +
                 str(fzl)+' m')

    att_method = dscfg.get('ATT_METHOD', 'ZPhi')
    if att_method not in ('ZPhi', 'Philin'):
        raise ValueError(
            'Unknown attenuation correction method. ' +
            'Must be one of the following: [ZPhi, Philin]')

    if att_method == 'ZPhi':
        spec_at, pia, cor_z, spec_diff_at, pida, cor_zdr = (
            pyart.correct.calculate_attenuation_zphi(
                radar, doc=15, fzl=fzl, smooth_window_len=0, a_coef=None,
                beta=None, c=None, d=None, refl_field=refl, phidp_field=phidp,
                zdr_field=zdr, temp_field=temp, iso0_field=iso0,
                spec_at_field=None, pia_field=None, corr_refl_field=None,
                spec_diff_at_field=None, pida_field=None, corr_zdr_field=None,
                temp_ref=temp_ref))
    elif att_method == 'Philin':
        spec_at, pia, cor_z, spec_diff_at, pida, cor_zdr = (
            pyart.correct.calculate_attenuation_philinear(
                radar, doc=15, fzl=fzl, pia_coef=None, pida_coef=None,
                refl_field=refl, phidp_field=phidp, zdr_field=zdr,
                temp_field=temp, iso0_field=iso0, spec_at_field=None,
                pia_field=None, corr_refl_field=None, spec_diff_at_field=None,
                pida_field=None, corr_zdr_field=None, temp_ref=temp_ref))

    # prepare for exit
    new_dataset = {'radar_out': deepcopy(radar)}
    new_dataset['radar_out'].fields = dict()

    new_dataset['radar_out'].add_field('specific_attenuation', spec_at)
    new_dataset['radar_out'].add_field('path_integrated_attenuation', pia)
    new_dataset['radar_out'].add_field('corrected_reflectivity', cor_z)

    if (spec_diff_at is not None) and (cor_zdr is not None):
        new_dataset['radar_out'].add_field(
            'specific_differential_attenuation', spec_diff_at)
        new_dataset['radar_out'].add_field(
            'path_integrated_differential_attenuation', pida)
        new_dataset['radar_out'].add_field(
            'corrected_differential_reflectivity', cor_zdr)
    else:
        warn(
            ' Specific differential attenuation and attenuation ' +
            'corrected differential reflectivity not available')

    return new_dataset, ind_rad



"""
pyrad.prod.process_timeseries_products
======================================

Functions for obtaining Pyrad products from a time series datasets

.. autosummary::
    :toctree: generated/

    generate_timeseries_products

"""

from copy import deepcopy
from warnings import warn

import numpy as np
from netCDF4 import num2date

from ..io.io_aux import get_save_dir, make_filename, get_fieldname_pyart
from ..io.io_aux import generate_field_name_str

from ..io.read_data_sensor import get_sensor_data
from ..io.read_data_other import read_timeseries

from ..io.write_data import write_ts_polar_data, write_ts_cum
from ..io.write_data import write_ts_grid_data, write_multiple_points

from ..graph.plots_timeseries import plot_timeseries, plot_timeseries_comp
from ..graph.plots_vol import plot_cappi, plot_traj
from ..graph.plots import plot_scatter_comp

from ..util.radar_utils import rainfall_accumulation


def generate_timeseries_products(dataset, prdcfg):
    """
    Generates time series products. Accepted product types:
        'COMPARE_CUMULATIVE_POINT': Plots in the same graph 2 time series of
            data accumulation (tipically rainfall rate). One time series is
            a point measurement of radar data while the other is from a
            co-located instrument (rain gauge or disdrometer)
            User defined parameters:
                dpi: int
                    The pixel density of the plot. Default 72
                vmin, vmax: float
                    The limits of the Y-axis. If none they will be obtained
                    from the Py-ART config file.
                sensor: str
                    The sensor type. Can be 'rgage' or 'disdro'
                sensorid: str
                    The sensor ID.
                location: str
                    A string identifying the location of the disdrometer
                freq: float
                    The frequency used to retrieve the polarimetric variables
                    of a disdrometer
                ele: float
                    The elevation angle used to retrieve the polarimetric
                    variables of a disdrometer
                ScanPeriod: float
                    The scaning period of the radar in seconds. This parameter
                    is defined in the 'loc' config file
        'COMPARE_POINT': Plots in the same graph 2 time series of
            data . One time series is a point measurement of radar data while
            the other is from a co-located instrument (rain gauge or
            disdrometer)
            User defined parameters:
                dpi: int
                    The pixel density of the plot. Default 72
                vmin, vmax: float
                    The limits of the Y-axis. If none they will be obtained
                    from the Py-ART config file.
                sensor: str
                    The sensor type. Can be 'rgage' or 'disdro'
                sensorid: str
                    The sensor ID.
                location: str
                    A string identifying the location of the disdrometer
                freq: float
                    The frequency used to retrieve the polarimetric variables
                    of a disdrometer
                ele: float
                    The elevation angle used to retrieve the polarimetric
                    variables of a disdrometer
        'COMPARE_TIME_AVG': Creates a scatter plot of average radar data
            versus average sensor data.
            User defined parameters:
                dpi: int
                    The pixel density of the plot. Default 72
                sensor: str
                    The sensor type. Can be 'rgage' or 'disdro'
                sensorid: str
                    The sensor ID.
                location: str
                    A string identifying the location of the disdrometer
                freq: float
                    The frequency used to retrieve the polarimetric variables
                    of a disdrometer
                ele: float
                    The elevation angle used to retrieve the polarimetric
                    variables of a disdrometer
                cum_time: float
                    Data accumulation time [s]. Default 3600.
                base_time: float
                    Starting moment of the accumulation [s from midnight].
                    Default 0.
        'PLOT_AND_WRITE': Writes and plots a trajectory time series.
            User defined parameters:
                ymin, ymax: float
                    The minimum and maximum value of the Y-axis. If none it
                    will be obtained from the Py-ART config file.
        'PLOT_AND_WRITE_POINT': Plots and writes a time series of radar data
            at a particular point
            User defined parameters:
                dpi: int
                    The pixel density of the plot. Default 72
                vmin, vmax: float
                    The limits of the Y-axis. If none they will be obtained
                    from the Py-ART config file.
        'PLOT_CUMULATIVE_POINT': Plots a time series of radar data
            accumulation at a particular point.
            User defined parameters:
                dpi: int
                    The pixel density of the plot. Default 72
                vmin, vmax: float
                    The limits of the Y-axis. If none they will be obtained
                    from the Py-ART config file.
                ScanPeriod: float
                    The scaning period of the radar in seconds. This parameter
                    is defined in the 'loc' config file
        'PLOT_HIST': plots and writes a histogram of all the data gathered
            during the trajectory processing
            User defined parameters:
                step: float or None
                    The quantization step of the data. If None it will be
                    obtained from the Py-ART config file
        'TRAJ_CAPPI_IMAGE': Creates a CAPPI image with the trajectory position
            overplot on it.
            User defined parameters:
                color_ref: str
                    The meaning of the color code with which the trajectory is
                    plotted. Can be 'None', 'altitude' (the absolute
                    altitude), 'rel_altitude' (altitude relative to the CAPPI
                    altitude), 'time' (trajectory time respect of the start of
                    the radar scan leading to the CAPPI)
                altitude: float
                    The CAPPI altitude [m]
                wfunc: str
                    Function used in the gridding of the radar data. The
                    function types are defined in pyart.map.grid_from_radars.
                    Default 'NEAREST'
                res: float
                    The CAPPI resolution [m]. Default 500.

    Parameters
    ----------
    dataset : dictionary
        radar object

    prdcfg : dictionary of dictionaries
        product configuration dictionary of dictionaries

    Returns
    -------
    no return

    """
    dssavedir = prdcfg['dsname']
    if 'dssavename' in prdcfg:
        dssavedir = prdcfg['dssavename']

    prdsavedir = prdcfg['prdname']
    if 'prdsavedir' in prdcfg:
        prdsavedir = prdcfg['prdsavedir']

    if prdcfg['type'] == 'WRITE_MULTIPLE_POINTS':

        savedir = get_save_dir(
            prdcfg['basepath'], prdcfg['procname'], dssavedir, prdsavedir,
            timeinfo=dataset['ref_time'])

        csvfname = make_filename(
            'POI', prdcfg['dstype'], dataset['datatype'], ['csv'],
            timeinfo=dataset['ref_time'])[0]

        csvfname = savedir+csvfname
        write_multiple_points(dataset, csvfname)
        print('saved CSV file: '+csvfname)

        return csvfname

    if prdcfg['type'] == 'PLOT_AND_WRITE_POINT':
        set_time_info = prdcfg.get('set_time_info', True)
        if 'antenna_coordinates_az_el_r' in dataset:
            az = '{:.1f}'.format(
                dataset['antenna_coordinates_az_el_r'][0])
            el = '{:.1f}'.format(
                dataset['antenna_coordinates_az_el_r'][1])
            r = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][2])
            gateinfo = ('az'+az+'r'+r+'el'+el)
        else:
            lon = '{:.3f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][0])
            lat = '{:.3f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][1])
            alt = '{:.1f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][2])
            gateinfo = ('lon'+lon+'lat'+lat+'alt'+alt)

        timeformat = None
        timeinfo = None
        if set_time_info:
            timeformat = '%Y%m%d'
            timeinfo = dataset['ref_time']

        savedir = get_save_dir(
            prdcfg['basepath'], prdcfg['procname'], dssavedir, prdsavedir,
            timeinfo=timeinfo)

        csvfname = make_filename(
            'ts', prdcfg['dstype'], dataset['datatype'], ['csv'],
            prdcfginfo=gateinfo, timeinfo=timeinfo,
            timeformat=timeformat)[0]

        csvfname = savedir+csvfname

        if not dataset['final']:
            if 'antenna_coordinates_az_el_r' in dataset:
                write_ts_polar_data(dataset, csvfname)
            else:
                write_ts_grid_data(dataset, csvfname)
            print('saved CSV file: '+csvfname)

        dpi = prdcfg.get('dpi', 72)
        vmin = prdcfg.get('vmin', None)
        vmax = prdcfg.get('vmax', None)
        plot_only_final = prdcfg.get('plot_only_final', False)

        if plot_only_final and not dataset['final']:
            return None
        if not plot_only_final and dataset['final']:
            return None

        date, value = read_timeseries(csvfname)
        if date is None:
            warn(
                'Unable to plot time series. No valid data')
            return None

        timeinfo_fig = None
        if set_time_info:
            timeinfo_fig = date[0]

        figfname_list = make_filename(
            'ts', prdcfg['dstype'], dataset['datatype'],
            prdcfg['imgformat'], prdcfginfo=gateinfo,
            timeinfo=timeinfo_fig, timeformat=timeformat)

        for i, figfname in enumerate(figfname_list):
            figfname_list[i] = savedir+figfname

        if 'antenna_coordinates_az_el_r' in dataset:
            label1 = 'Radar (az, el, r): ('+az+', '+el+', '+r+')'
        else:
            label1 = 'Grid (lon, lat, alt): ('+lon+', '+lat+', '+alt+')'
        titl = ('Time Series '+date[0].strftime('%Y-%m-%d'))

        labely = generate_field_name_str(dataset['datatype'])

        plot_timeseries(
            date, [value], figfname_list, labelx='Time UTC',
            labely=labely, labels=[label1], title=titl, dpi=dpi,
            ymin=vmin, ymax=vmax)
        print('----- save to '+' '.join(figfname_list))

        return figfname_list

    if prdcfg['type'] == 'PLOT_CUMULATIVE_POINT':
        dpi = prdcfg.get('dpi', 72)
        vmin = prdcfg.get('vmin', None)
        vmax = prdcfg.get('vmax', None)
        plot_only_final = prdcfg.get('plot_only_final', False)
        set_time_info = prdcfg.get('set_time_info', True)

        if plot_only_final and not dataset['final']:
            return None
        if not plot_only_final and dataset['final']:
            return None

        if 'antenna_coordinates_az_el_r' in dataset:
            az = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][0])
            el = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][1])
            r = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][2])
            gateinfo = ('az'+az+'r'+r+'el'+el)
        else:
            lon = '{:.3f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][0])
            lat = '{:.3f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][1])
            alt = '{:.1f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][2])
            gateinfo = ('lon'+lon+'lat'+lat+'alt'+alt)

        timeformat = None
        timeinfo = None
        if set_time_info:
            timeformat = '%Y%m%d'
            timeinfo = dataset['ref_time']

        savedir = get_save_dir(
            prdcfg['basepath'], prdcfg['procname'], dssavedir,
            prdcfg['prdid'], timeinfo=timeinfo)

        csvfname = make_filename(
            'ts', prdcfg['dstype'], dataset['datatype'], ['csv'],
            prdcfginfo=gateinfo, timeinfo=timeinfo,
            timeformat=timeformat)[0]

        csvfname = savedir+csvfname

        date, value = read_timeseries(csvfname)

        if date is None:
            warn(
                'Unable to plot accumulation time series. No valid data')
            return None

        timeinfo_fig = None
        if set_time_info:
            timeinfo_fig = date[0]

        figfname_list = make_filename(
            'ts_cum', prdcfg['dstype'], dataset['datatype'],
            prdcfg['imgformat'], prdcfginfo=gateinfo,
            timeinfo=timeinfo_fig, timeformat=timeformat)

        for i, figfname in enumerate(figfname_list):
            figfname_list[i] = savedir+figfname

        if 'antenna_coordinates_az_el_r' in dataset:
            label1 = 'Radar (az, el, r): ('+az+', '+el+', '+r+')'
        else:
            label1 = 'Grid (lon, lat, alt): ('+lon+', '+lat+', '+alt+')'
        titl = ('Time Series Acc. '+date[0].strftime('%Y-%m-%d'))

        labely = 'Radar estimated rainfall accumulation (mm)'

        plot_timeseries(
            date, [value], figfname_list, labelx='Time UTC',
            labely=labely, labels=[label1], title=titl,
            period=prdcfg['ScanPeriod']*60.,
            ymin=vmin, ymax=vmax, dpi=dpi)
        print('----- save to '+' '.join(figfname_list))

        return figfname_list

    if prdcfg['type'] == 'COMPARE_POINT':
        dpi = prdcfg.get('dpi', 72)
        vmin = prdcfg.get('vmin', None)
        vmax = prdcfg.get('vmax', None)
        plot_only_final = prdcfg.get('plot_only_final', False)
        set_time_info = prdcfg.get('set_time_info', True)

        if plot_only_final and not dataset['final']:
            return None
        if not plot_only_final and dataset['final']:
            return None

        if 'antenna_coordinates_az_el_r' in dataset:
            az = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][0])
            el = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][1])
            r = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][2])
            gateinfo = ('az'+az+'r'+r+'el'+el)
        else:
            lon = '{:.3f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][0])
            lat = '{:.3f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][1])
            alt = '{:.1f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][2])
            gateinfo = ('lon'+lon+'lat'+lat+'alt'+alt)

        timeformat = None
        timeinfo = None
        if set_time_info:
            timeformat = '%Y%m%d'
            timeinfo = dataset['ref_time']

        savedir_ts = get_save_dir(
            prdcfg['basepath'], prdcfg['procname'], dssavedir,
            prdcfg['prdid'], timeinfo=timeinfo)

        csvfname = make_filename(
            'ts', prdcfg['dstype'], dataset['datatype'], ['csv'],
            prdcfginfo=gateinfo, timeinfo=timeinfo,
            timeformat=timeformat)[0]

        csvfname = savedir_ts+csvfname

        radardate, radarvalue = read_timeseries(csvfname)
        if radardate is None:
            warn(
                'Unable to plot sensor comparison at point of interest. ' +
                'No valid radar data')
            return None

        sensordate, sensorvalue, sensortype, _ = get_sensor_data(
            radardate[0], dataset['datatype'], prdcfg)
        if sensordate is None:
            warn(
                'Unable to plot sensor comparison at point of interest. ' +
                'No valid sensor data')
            return None

        timeinfo_fig = None
        if set_time_info:
            timeinfo_fig = radardate[0]

        savedir = get_save_dir(
            prdcfg['basepath'], prdcfg['procname'], dssavedir,
            prdsavedir, timeinfo=timeinfo_fig)

        figfname_list = make_filename(
            'ts_comp', prdcfg['dstype'], dataset['datatype'],
            prdcfg['imgformat'], prdcfginfo=gateinfo,
            timeinfo=timeinfo_fig, timeformat=timeformat)

        for i, figfname in enumerate(figfname_list):
            figfname_list[i] = savedir+figfname

        if 'antenna_coordinates_az_el_r' in dataset:
            label1 = 'Radar (az, el, r): ('+az+', '+el+', '+r+')'
        else:
            label1 = 'Grid (lon, lat, alt): ('+lon+', '+lat+', '+alt+')'
        label2 = sensortype+' '+prdcfg['sensorid']
        titl = 'Time Series Comp. '+radardate[0].strftime('%Y-%m-%d')
        labely = generate_field_name_str(dataset['datatype'])

        plot_timeseries_comp(
            radardate, radarvalue, sensordate, sensorvalue, figfname_list,
            labelx='Time UTC', labely=labely, label1=label1, label2=label2,
            titl=titl, ymin=vmin, ymax=vmax, dpi=dpi)
        print('----- save to '+' '.join(figfname_list))

        return figfname_list

    if prdcfg['type'] == 'COMPARE_CUMULATIVE_POINT':
        dpi = prdcfg.get('dpi', 72)
        vmin = prdcfg.get('vmin', None)
        vmax = prdcfg.get('vmax', None)
        plot_only_final = prdcfg.get('plot_only_final', False)
        set_time_info = prdcfg.get('set_time_info', True)

        if plot_only_final and not dataset['final']:
            return None
        if not plot_only_final and dataset['final']:
            return None

        if 'antenna_coordinates_az_el_r' in dataset:
            az = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][0])
            el = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][1])
            r = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][2])
            gateinfo = ('az'+az+'r'+r+'el'+el)
        else:
            lon = '{:.3f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][0])
            lat = '{:.3f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][1])
            alt = '{:.1f}'.format(
                dataset['point_coordinates_WGS84_lon_lat_alt'][2])
            gateinfo = ('lon'+lon+'lat'+lat+'alt'+alt)

        timeformat = None
        timeinfo = None
        if set_time_info:
            timeformat = '%Y%m%d'
            timeinfo = dataset['ref_time']

        savedir_ts = get_save_dir(
            prdcfg['basepath'], prdcfg['procname'], dssavedir,
            prdcfg['prdid'], timeinfo=timeinfo)

        csvfname = make_filename(
            'ts', prdcfg['dstype'], dataset['datatype'], ['csv'],
            prdcfginfo=gateinfo, timeinfo=timeinfo,
            timeformat=timeformat)[0]

        csvfname = savedir_ts+csvfname

        radardate, radarvalue = read_timeseries(csvfname)
        if radardate is None:
            warn(
                'Unable to plot sensor comparison at point of interest. ' +
                'No valid radar data')
            return None

        sensordate, sensorvalue, sensortype, period2 = get_sensor_data(
            radardate[0], dataset['datatype'], prdcfg)
        if sensordate is None:
            warn(
                'Unable to plot sensor comparison at point of interest. ' +
                'No valid sensor data')
            return None

        timeinfo_fig = None
        if set_time_info:
            timeinfo_fig = radardate[0]

        savedir = get_save_dir(
            prdcfg['basepath'], prdcfg['procname'], dssavedir,
            prdsavedir, timeinfo=timeinfo_fig)

        figfname_list = make_filename(
            'ts_cumcomp', prdcfg['dstype'], dataset['datatype'],
            prdcfg['imgformat'], prdcfginfo=gateinfo,
            timeinfo=timeinfo_fig, timeformat=timeformat)

        for i, figfname in enumerate(figfname_list):
            figfname_list[i] = savedir+figfname

        if 'antenna_coordinates_az_el_r' in dataset:
            label1 = 'Radar (az, el, r): ('+az+', '+el+', '+r+')'
        else:
            label1 = 'Grid (lon, lat, alt): ('+lon+', '+lat+', '+alt+')'
        label2 = sensortype+' '+prdcfg['sensorid']
        titl = ('Time Series Acc. Comp. ' +
                radardate[0].strftime('%Y-%m-%d'))
        labely = 'Rainfall accumulation (mm)'

        plot_timeseries_comp(
            radardate, radarvalue, sensordate, sensorvalue,
            figfname_list, labelx='Time UTC', labely=labely,
            label1=label1, label2=label2, titl=titl,
            period1=prdcfg['ScanPeriod']*60., period2=period2,
            ymin=vmin, ymax=vmax, dpi=dpi)
        print('----- save to '+' '.join(figfname_list))

        return figfname_list

    if prdcfg['type'] == 'COMPARE_TIME_AVG':
        dpi = prdcfg.get('dpi', 72)
        plot_only_final = prdcfg.get('plot_only_final', False)

        if plot_only_final and not dataset['final']:
            return None
        if not plot_only_final and dataset['final']:
            return None

        az = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][0])
        el = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][1])
        r = '{:.1f}'.format(dataset['antenna_coordinates_az_el_r'][2])
        gateinfo = ('az'+az+'r'+r+'el'+el)

        savedir_ts = get_save_dir(
            prdcfg['basepath'], prdcfg['procname'], dssavedir,
            prdcfg['prdid'], timeinfo=dataset['time'])

        csvfname = make_filename(
            'ts', prdcfg['dstype'], dataset['datatype'], ['csv'],
            prdcfginfo=gateinfo, timeinfo=dataset['time'],
            timeformat='%Y%m%d')[0]

        radardate, radarvalue = read_timeseries(savedir_ts+csvfname)
        if radardate is None:
            warn(
                'Unable to compared time averaged data at POI. ' +
                'No valid radar data')
            return None

        sensordate, sensorvalue, sensortype, period2 = get_sensor_data(
            radardate[0], dataset['datatype'], prdcfg)
        if sensordate is None:
            warn(
                'Unable to compared time averaged data at POI. ' +
                'No valid sensor data')
            return None

        cum_time = prdcfg.get('cum_time', 3600)
        base_time = prdcfg.get('base_time', 0)

        sensordate_cum, sensorvalue_cum, np_sensor_cum = rainfall_accumulation(
            sensordate, sensorvalue, cum_time=cum_time, base_time=base_time,
            dropnan=False)

        radardate_cum, radarvalue_cum, np_radar_cum = rainfall_accumulation(
            radardate, radarvalue, cum_time=cum_time, base_time=base_time,
            dropnan=False)

        # find common time stamps
        ind = np.where(np.in1d(radardate_cum, sensordate_cum))[0]
        if ind.size == 0:
            warn('No sensor data for radar data time stamps')
        radardate_cum2 = radardate_cum[ind]
        radarvalue_cum2 = radarvalue_cum[ind]
        np_radar_cum2 = np_radar_cum[ind]

        ind = np.where(np.in1d(sensordate_cum, radardate_cum2))[0]
        sensorvalue_cum2 = sensorvalue_cum[ind]
        np_sensor_cum2 = np_sensor_cum[ind]

        savedir = get_save_dir(
            prdcfg['basepath'], prdcfg['procname'], dssavedir,
            prdsavedir, timeinfo=radardate[0])

        fname = make_filename(
            str(cum_time)+'s_acc_ts_comp', prdcfg['dstype'],
            dataset['datatype'], ['csv'], prdcfginfo=gateinfo,
            timeinfo=radardate[0], timeformat='%Y%m%d')[0]

        fname = savedir+fname

        new_dataset = deepcopy(dataset)
        new_dataset.update({
            'time': radardate_cum2,
            'sensor_value': sensorvalue_cum2,
            'np_sensor': np_sensor_cum2,
            'radar_value': radarvalue_cum2,
            'np_radar': np_radar_cum2,
            'cum_time': cum_time})
        new_dataset.update(prdcfg)

        write_ts_cum(new_dataset, fname)

        print('saved CSV file: '+fname)

        figfname_list = make_filename(
            str(cum_time)+'s_acc_ts_comp', prdcfg['dstype'],
            dataset['datatype'], prdcfg['imgformat'], prdcfginfo=gateinfo,
            timeinfo=radardate[0], timeformat='%Y%m%d')

        for i, figfname in enumerate(figfname_list):
            figfname_list[i] = savedir+figfname

        labelx = sensortype+' '+prdcfg['sensorid']+' (mm)'
        labely = 'Radar (az, el, r): ('+az+', '+el+', '+r+') (mm)'
        titl = (str(cum_time)+'s Acc. Comp. ' +
                radardate_cum[0].strftime('%Y-%m-%d'))

        plot_scatter_comp(
            sensorvalue_cum2, radarvalue_cum2, figfname_list, labelx=labelx,
            labely=labely, titl=titl, axis='equal', dpi=dpi)

        print('----- save to '+' '.join(figfname_list))

        return figfname_list

    # ================================================================
    if prdcfg['type'] == 'PLOT_AND_WRITE':
        if not dataset['final']:
            return None

        field_name = get_fieldname_pyart(prdcfg['voltype'])
        if field_name not in dataset['ts_dict']:
            warn(
                ' Field type ' + field_name +
                ' not available in data set. Skipping product ' +
                prdcfg['type'])
            return None

        ts = dataset['ts_dict'][field_name]
        timeinfo = ts.time_vector[0]

        savedir = get_save_dir(prdcfg['basepath'], prdcfg['procname'],
                               dssavedir, prdsavedir,
                               timeinfo=timeinfo)

        dstype_str = prdcfg['dstype'].lower().replace('_', '')
        fname = make_filename('ts', dstype_str, ts.datatype,
                              ['csv'],
                              prdcfginfo=None, timeinfo=timeinfo,
                              timeformat='%Y%m%d%H%M%S',
                              runinfo=prdcfg['runinfo'])

        ts.write(savedir + fname[0])

        fname = make_filename('ts', dstype_str, ts.datatype,
                              prdcfg['imgformat'],
                              prdcfginfo=None, timeinfo=timeinfo,
                              timeformat='%Y%m%d%H%M%S',
                              runinfo=prdcfg['runinfo'])

        ymin = prdcfg.get('ymin', None)
        ymax = prdcfg.get('ymax', None)

        ts.plot(savedir + fname[0], ymin=ymin, ymax=ymax)

        return None

    if prdcfg['type'] == 'PLOT_HIST':
        if not dataset['final']:
            return None

        field_name = get_fieldname_pyart(prdcfg['voltype'])
        if field_name not in dataset['ts_dict']:
            warn(
                ' Field type ' + field_name +
                ' not available in data set. Skipping product ' +
                prdcfg['type'])
            return None

        ts = dataset['ts_dict'][field_name]
        timeinfo = ts.time_vector[0]

        savedir = get_save_dir(prdcfg['basepath'], prdcfg['procname'],
                               dssavedir, prdsavedir,
                               timeinfo=timeinfo)

        dstype_str = prdcfg['dstype'].lower().replace('_', '')

        fname = make_filename('hist', dstype_str, ts.datatype,
                              prdcfg['imgformat'],
                              prdcfginfo=None, timeinfo=timeinfo,
                              timeformat='%Y%m%d%H%M%S',
                              runinfo=prdcfg['runinfo'])

        step = prdcfg.get('step', None)

        ts.plot_hist(savedir + fname[0], step=step)

        return None

    if prdcfg['type'] == 'TRAJ_CAPPI_IMAGE':
        if dataset['final']:
            return None

        field_name = get_fieldname_pyart(prdcfg['voltype'])
        if field_name not in dataset['radar'].fields:
            warn(
                ' Field type ' + field_name +
                ' not available in data set. Skipping product ' +
                prdcfg['type'])
            return None

        savedir = get_save_dir(prdcfg['basepath'], prdcfg['procname'],
                               dssavedir, prdsavedir,
                               timeinfo=prdcfg['timeinfo'])

        color_ref = prdcfg.get('color_ref', 'None')
        if color_ref == 'altitude':
            prdtype = 'cappi_alt'
        elif color_ref == 'rel_altitude':
            prdtype = 'cappi_rel_alt'
        elif color_ref == 'time':
            prdtype = 'cappi_time'
        else:
            prdtype = 'cappi'

        fname_list = make_filename(
            prdtype, prdcfg['dstype'], prdcfg['voltype'],
            prdcfg['imgformat'],
            prdcfginfo='alt'+'{:.1f}'.format(prdcfg['altitude']),
            timeinfo=prdcfg['timeinfo'], runinfo=prdcfg['runinfo'])

        for i, fname in enumerate(fname_list):
            fname_list[i] = savedir+fname

        fig, ax = plot_cappi(
            dataset['radar'], field_name, prdcfg['altitude'], prdcfg,
            fname_list, save_fig=False)

        # start time of radar object
        t_start = dataset['radar'].time['data'].min()
        dt_start = num2date(t_start, dataset['radar'].time['units'],
                            dataset['radar'].time['calendar'])

        fname_list = plot_traj(
            dataset['rng_traj'], dataset['azi_traj'], dataset['ele_traj'],
            dataset['time_traj'], prdcfg, fname_list,
            rad_alt=dataset['radar'].altitude['data'], rad_tstart=dt_start,
            ax=ax, fig=fig, save_fig=True)

        print('----- saved to '+' '.join(fname_list))

        return None

    # ================================================================
    warn(' Unsupported product type: ' + prdcfg['type'])
    return None



"""
pyrad.proc.process_iq
=====================

Functions to processes IQ data.

.. autosummary::
    :toctree: generated/

    process_raw_iq
    process_pol_variables_iq
    process_reflectivity_iq
    process_st1_iq
    process_st2_iq
    process_wbn_iq
    process_differential_reflectivity_iq
    process_mean_phase_iq
    process_differential_phase_iq
    process_rhohv_iq
    process_Doppler_velocity_iq
    process_Doppler_width_iq
    process_fft

"""

from copy import deepcopy
from warnings import warn
import numpy as np
from netCDF4 import num2date

import pyart

from ..io.io_aux import get_datatype_fields, get_fieldname_pyart


def process_raw_iq(procstatus, dscfg, radar_list=None):
    """
    Dummy function that returns the initial input data set

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg['datatype']:
        radarnr, _, _, _, _ = get_datatype_fields(datatypedescr)
        break
    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    new_dataset = {'radar_out': deepcopy(radar_list[ind_rad])}

    return new_dataset, ind_rad


def process_pol_variables_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the polarimetric variables from the IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
        subtract_noise : Bool
            If True noise will be subtracted from the signal
        lag : int
            The time lag to use in the estimators
        direction : str
            The convention used in the Doppler mean field. Can be
            negative_away or negative_towards
        variables : list of str
            list of variables to compute. Default dBZ
        phase_offset : float. Dataset keyword
            The system differential phase offset to remove
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    noise_h_field = None
    noise_v_field = None
    signal_h_field = None
    signal_v_field = None
    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'IQhhADU':
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype == 'IQvvADU':
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype == 'IQNADUh':
            noise_h_field = get_fieldname_pyart(datatype)
        elif datatype == 'IQNADUv':
            noise_v_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    subtract_noise = dscfg.get('subtract_noise', False)
    lag = dscfg.get('lag', None)
    direction = dscfg.get('direction', 'negative_away')
    variables = dscfg.get('variables', ['dBZ'])
    phase_offset = dscfg.get('phase_offset', 0.)

    fields_list = []
    for variable in variables:
        fields_list.append(get_fieldname_pyart(variable))

    radar = pyart.retrieve.compute_pol_variables_iq(
        radar, fields_list, subtract_noise=subtract_noise, lag=lag,
        direction=direction, phase_offset=phase_offset,
        signal_h_field=signal_h_field, signal_v_field=signal_v_field,
        noise_h_field=noise_h_field, noise_v_field=noise_v_field)

    # prepare for exit
    new_dataset = {'radar_out': radar}

    return new_dataset, ind_rad


def process_reflectivity_iq(procstatus, dscfg, radar_list=None):
    """
    Computes reflectivity from the IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
        subtract_noise : Bool
            If True noise will be subtracted from the signal
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    noise_field = None
    signal_field = None
    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ('IQhhADU', 'IQvvADU'):
            signal_field = get_fieldname_pyart(datatype)
        elif datatype in ('IQNADUh', 'IQNADUv'):
            noise_field = get_fieldname_pyart(datatype)

    if signal_field is None:
        warn('Signal field must be specified')
        return None, None

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn('Unable to obtain reflectivity. Missing field ' +
             signal_field)
        return None, None

    subtract_noise = dscfg.get('subtract_noise', False)

    dBZ = pyart.retrieve.compute_reflectivity_iq(
        radar, subtract_noise=subtract_noise, signal_field=signal_field,
        noise_field=noise_field)

    reflectivity_field = 'reflectivity'
    if signal_field in ('IQ_vv_ADU',):
        reflectivity_field += '_vv'

    # prepare for exit
    new_dataset = {'radar_out': pyart.util.radar_from_spectra(radar)}
    new_dataset['radar_out'].add_field(reflectivity_field, dBZ)

    return new_dataset, ind_rad


def process_st1_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the statistical test one lag fluctuation from the horizontal or
    vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg['datatype'][0])
    signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn('Unable to obtain ST1. Missing fields')
        return None, None

    st1 = pyart.retrieve.compute_st1_iq(
        radar, signal_field=signal_field)

    st1_field = 'stat_test_lag1'
    if signal_field == 'IQ_vv_ADU':
        st1_field += '_vv'

    # prepare for exit
    new_dataset = {'radar_out': pyart.util.radar_from_spectra(radar)}
    new_dataset['radar_out'].add_field(st1_field, st1)

    return new_dataset, ind_rad


def process_st2_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the statistical test two lag fluctuation from the horizontal or
    vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg['datatype'][0])
    signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn('Unable to obtain ST2. Missing fields')
        return None, None

    st2 = pyart.retrieve.compute_st2_iq(
        radar, signal_field=signal_field)

    st2_field = 'stat_test_lag2'
    if signal_field == 'IQ_vv_ADU':
        st2_field += '_vv'

    # prepare for exit
    new_dataset = {'radar_out': pyart.util.radar_from_spectra(radar)}
    new_dataset['radar_out'].add_field(st2_field, st2)

    return new_dataset, ind_rad


def process_wbn_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the wide band noise from the horizontal or vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg['datatype'][0])
    signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn('Unable to obtain WBN. Missing fields')
        return None, None

    wbn = pyart.retrieve.compute_wbn_iq(
        radar, signal_field=signal_field)

    wbn_field = 'wide_band_noise'
    if signal_field == 'IQ_vv_ADU':
        wbn_field += '_vv'

    # prepare for exit
    new_dataset = {'radar_out': pyart.util.radar_from_spectra(radar)}
    new_dataset['radar_out'].add_field(wbn_field, wbn)

    return new_dataset, ind_rad


def process_differential_reflectivity_iq(procstatus, dscfg, radar_list=None):
    """
    Computes differential reflectivity from the horizontal and vertical
    IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
        subtract_noise : Bool
            If True noise will be subtracted from the signal
        lag : int
            The time lag to use in the estimators
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    noise_h_field = None
    noise_v_field = None
    signal_h_field = None
    signal_v_field = None
    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'IQhhADU':
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype == 'IQvvADU':
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype == 'IQNADUh':
            noise_h_field = get_fieldname_pyart(datatype)
        elif datatype == 'IQNADUv':
            noise_v_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if (signal_h_field not in radar.fields or
            signal_v_field not in radar.fields):
        warn('Unable to obtain spectral differential reflectivity. ' +
             'Missing fields')
        return None, None

    subtract_noise = dscfg.get('subtract_noise', False)
    lag = dscfg.get('lag', 0)

    zdr = pyart.retrieve.compute_differential_reflectivity_iq(
        radar, subtract_noise=subtract_noise, lag=lag,
        signal_h_field=signal_h_field, signal_v_field=signal_v_field,
        noise_h_field=noise_h_field, noise_v_field=noise_v_field)

    # prepare for exit
    new_dataset = {'radar_out': pyart.util.radar_from_spectra(radar)}
    new_dataset['radar_out'].add_field('differential_reflectivity', zdr)

    return new_dataset, ind_rad


def process_mean_phase_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the mean phase from the horizontal or vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg['datatype'][0])
    signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn('Unable to obtain MPH. Missing fields')
        return None, None

    mph = pyart.retrieve.compute_mean_phase_iq(
        radar, signal_field=signal_field)

    mean_phase_field = 'mean_phase'
    if signal_field == 'IQ_vv_ADU':
        mean_phase_field += '_vv'

    # prepare for exit
    new_dataset = {'radar_out': pyart.util.radar_from_spectra(radar)}
    new_dataset['radar_out'].add_field(mean_phase_field, mph)

    return new_dataset, ind_rad


def process_differential_phase_iq(procstatus, dscfg, radar_list=None):
    """
    Computes the differential phase from the horizontal and vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
        phase_offset : float. Dataset keyword
            The system differential phase offset to remove
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    signal_h_field = None
    signal_v_field = None
    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'IQhhADU':
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype == 'IQvvADU':
            signal_v_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if (signal_h_field not in radar.fields or
            signal_v_field not in radar.fields):
        warn('Unable to obtain PhiDP. Missing fields')
        return None, None

    phase_offset = dscfg.get('phase_offset', 0.)

    uphidp = pyart.retrieve.compute_differential_phase_iq(
        radar, phase_offset=phase_offset, signal_h_field=signal_h_field,
        signal_v_field=signal_v_field)

    # prepare for exit
    new_dataset = {'radar_out': pyart.util.radar_from_spectra(radar)}
    new_dataset['radar_out'].add_field(
        'uncorrected_differential_phase', uphidp)

    return new_dataset, ind_rad


def process_rhohv_iq(procstatus, dscfg, radar_list=None):
    """
    Computes RhoHV from the horizontal and vertical IQ data

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
        subtract_noise : Bool
            If True noise will be subtracted from the signal
        lag : int
            Time lag used in the computation
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """
    if procstatus != 1:
        return None, None

    noise_h_field = None
    noise_v_field = None
    signal_h_field = None
    signal_v_field = None
    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype == 'IQhhADU':
            signal_h_field = get_fieldname_pyart(datatype)
        elif datatype == 'IQvvADU':
            signal_v_field = get_fieldname_pyart(datatype)
        elif datatype == 'IQNADUh':
            noise_h_field = get_fieldname_pyart(datatype)
        elif datatype == 'IQNADUv':
            noise_v_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if (signal_h_field not in radar.fields or
            signal_v_field not in radar.fields):
        warn('Unable to obtain RhoHV. Missing fields')
        return None, None

    subtract_noise = dscfg.get('subtract_noise', False)
    lag = dscfg.get('lag', 0)

    rhohv = pyart.retrieve.compute_rhohv_iq(
        radar, subtract_noise=subtract_noise, lag=lag,
        signal_h_field=signal_h_field, signal_v_field=signal_v_field,
        noise_h_field=noise_h_field, noise_v_field=noise_v_field)

    # prepare for exit
    new_dataset = {'radar_out': pyart.util.radar_from_spectra(radar)}
    new_dataset['radar_out'].add_field('cross_correlation_ratio', rhohv)

    return new_dataset, ind_rad


def process_Doppler_velocity_iq(procstatus, dscfg, radar_list=None):
    """
    Compute the Doppler velocity from the spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
        direction : str
            The convention used in the Doppler mean field. Can be
            negative_away or negative_towards
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ('IQhhADU', 'IQvvADU'):
            signal_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn('Unable to obtain Doppler velocity. ' +
             'Missing field '+signal_field)
        return None, None

    direction = dscfg.get('direction', 'negative_away')

    vel = pyart.retrieve.compute_Doppler_velocity_iq(
        radar, signal_field=signal_field, direction=direction)

    vel_field = 'velocity'
    if signal_field == 'IQ_vv_ADU':
        vel_field += '_vv'

    # prepare for exit
    new_dataset = {'radar_out': pyart.util.radar_from_spectra(radar)}
    new_dataset['radar_out'].add_field(vel_field, vel)

    return new_dataset, ind_rad


def process_Doppler_width_iq(procstatus, dscfg, radar_list=None):
    """
    Compute the Doppler spectrum width from the spectral reflectivity

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
        subtract_noise : Bool
            If True noise will be subtracted from the signals
        lag : int
            Time lag used in the denominator of the computation
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    noise_field = None
    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        if datatype in ('IQhhADU', 'IQvvADU'):
            signal_field = get_fieldname_pyart(datatype)
        elif datatype in ('IQNADUh', 'IQNADUv'):
            noise_field = get_fieldname_pyart(datatype)

    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    if signal_field not in radar.fields:
        warn('Unable to obtain Doppler spectrum width. ' +
             'Missing field '+signal_field)
        return None, None

    subtract_noise = dscfg.get('subtract_noise', False)
    lag = dscfg.get('lag', 1)

    width = pyart.retrieve.compute_Doppler_width_iq(
        radar, subtract_noise=True, signal_field=signal_field,
        noise_field=noise_field, lag=lag)

    width_field = 'spectrum_width'
    if signal_field == 'IQ_vv_ADU':
        width_field += '_vv'

    # prepare for exit
    new_dataset = {'radar_out': pyart.util.radar_from_spectra(radar)}
    new_dataset['radar_out'].add_field(width_field, width)

    return new_dataset, ind_rad


def process_fft(procstatus, dscfg, radar_list=None):
    """
    Compute the Doppler spectra form the IQ data with a Fourier transform

    Parameters
    ----------
    procstatus : int
        Processing status: 0 initializing, 1 processing volume,
        2 post-processing
    dscfg : dictionary of dictionaries
        data set configuration. Accepted configuration keywords::

        datatype : list of string. Dataset keyword
            The input data types
        window : list of str
            Parameters of the window used to obtain the spectra. The
            parameters are the ones corresponding to function
            scipy.signal.windows.get_window. It can also be ['None'].
    radar_list : list of spectra objects
        Optional. list of spectra objects

    Returns
    -------
    new_dataset : dict
        dictionary containing the output
    ind_rad : int
        radar index

    """

    if procstatus != 1:
        return None, None

    radarnr, _, datatype, _, _ = get_datatype_fields(dscfg['datatype'][0])
    ind_rad = int(radarnr[5:8])-1
    if (radar_list is None) or (radar_list[ind_rad] is None):
        warn('ERROR: No valid radar')
        return None, None
    radar = radar_list[ind_rad]

    wind_params = dscfg.get('window', ['None'])
    if len(wind_params) == 1:
        window = wind_params[0]
        if window == 'None':
            window = None
        else:
            try:
                window = float(window)
            except ValueError:
                pass
    else:
        window = wind_params
        for i in range(1, len(window)):
            window[i] = float(window[i])
        window = tuple(window)

    fields_in_list = []
    fields_out_list = []
    for datatypedescr in dscfg['datatype']:
        radarnr, _, datatype, _, _ = get_datatype_fields(datatypedescr)
        field_name = get_fieldname_pyart(datatype)
        if field_name not in radar.fields:
            warn(field_name+' not in radar')
            continue
        if field_name == 'IQ_hh_ADU':
            fields_out_list.append('unfiltered_complex_spectra_hh_ADU')
        elif field_name == 'IQ_vv_ADU':
            fields_out_list.append('unfiltered_complex_spectra_vv_ADU')
        elif field_name == 'IQ_noise_power_hh_ADU':
            fields_out_list.append('spectral_noise_power_hh_ADU')
        elif field_name == 'IQ_noiseADU_vv':
            fields_out_list.append('spectral_noise_power_vv_ADU')
        else:
            warn(field_name+' can not be Fourier transformed')
        fields_in_list.append(field_name)

    radar_out = pyart.retrieve.compute_spectra(
        radar, fields_in_list, fields_out_list, window=window)

    # prepare for exit
    new_dataset = {'radar_out': radar_out}

    return new_dataset, ind_rad




# THIS FILE IS GENERATED FROM PYRAD_PROC SETUP.PY
short_version = '1.1.2'
version = '1.1.2'
full_version = '1.1.2'
git_revision = '1adcbe285823ae3d8d718aab42126087dc6d3f98'
compile_date_time = '2022-05-04 10:41'
username = 'wolfensb'
release = True

if not release:
    version = full_version



"""
==================================
Utilities (:mod:`pyrad.util`)
==================================

.. currentmodule:: pyrad.util

Functions to read and write data and configuration files.

Radar Utilities
===============

.. autosummary::
    :toctree: generated/

    get_data_along_rng
    get_data_along_azi
    get_data_along_ele
    get_ROI
    rainfall_accumulation
    time_series_statistics
    find_contiguous_times
    join_time_series
    get_range_bins_to_avg
    find_ray_index
    find_rng_index
    find_nearest_gate
    find_colocated_indexes
    get_fixed_rng_data
    time_avg_range
    get_closest_solar_flux
    create_sun_hits_field
    create_sun_retrieval_field
    compute_quantiles
    compute_quantiles_from_hist
    compute_quantiles_sweep
    compute_2d_hist
    compute_1d_stats
    compute_2d_stats
    compute_histogram
    compute_histogram_sweep
    belongs_roi_indices
    get_cercle_coords
    compute_profile_stats
    project_to_vertical

    quantiles_weighted
    ratio_bootstrapping
"""

from .radar_utils import time_avg_range, get_closest_solar_flux
from .radar_utils import create_sun_hits_field, create_sun_retrieval_field
from .radar_utils import compute_histogram, compute_histogram_sweep
from .radar_utils import compute_quantiles, compute_quantiles_sweep
from .radar_utils import compute_quantiles_from_hist, get_range_bins_to_avg
from .radar_utils import find_ray_index, find_rng_index, find_nearest_gate
from .radar_utils import find_colocated_indexes, find_contiguous_times
from .radar_utils import compute_2d_hist, compute_1d_stats, compute_2d_stats
from .radar_utils import time_series_statistics, join_time_series
from .radar_utils import rainfall_accumulation, get_ROI, belongs_roi_indices
from .radar_utils import project_to_vertical, get_data_along_rng
from .radar_utils import get_data_along_azi, get_data_along_ele
from .radar_utils import get_fixed_rng_data, get_cercle_coords
from .radar_utils import compute_profile_stats

from .stat_utils import quantiles_weighted, ratio_bootstrapping

__all__ = [s for s in dir() if not s.startswith('_')]



"""
pyrad.io.config
===============

Functions for reading pyrad config files

.. autosummary::
    :toctree: generated/

    read_config
    get_num_elements
    string_to_datatype
    get_array
    get_struct
    get_array_type
    init_array

"""

import os
import re
import numpy as np


def read_config(fname, cfg=None):
    """
    Read a pyrad config file.

    Parameters
    ----------
    fname : str
        Name of the configuration file to read.

    cfg : dict of dicts, optional
        dictionary of dictionaries containing configuration parameters where
        the new parameters will be placed

    Returns
    -------
    cfg : dict of dicts
        dictionary of dictionaries containing the configuration parameters

    """

    # check if the file can be read
    try:
        cfgfile = open(fname, "r", encoding='utf-8', errors='ignore')
    except:
        raise Exception("ERROR: Could not find|open config file '"+fname+"'")

    # if config dictionary does not exist yet create it
    if cfg is None:
        cfg = dict()

    # read file contents
    fileend = 0
    while fileend == 0:
        # remove leading and trailing whitespace
        line = cfgfile.readline()
        if line:
            line = line.strip()

            # ignore white lines
            if not line:
                continue

            # ignore comments
            if line.startswith('#'):
                continue

            line = line.partition('#')[0]  # Remove comments
            line = line.strip()

            vals = line.split()

            fieldname = vals[0]
            nvals = len(vals)

            if nvals < 3:
                raise Exception(
                    "FILE FORMAT ERROR: file: " + fname +
                    ", variable: " + fieldname +
                    ": Wrong number of elements!")

            valtype = vals[1]
            valuestr = vals[2:nvals]
            nel, isstruct = get_num_elements(valtype, valuestr)

            if nel > 0:
                if isstruct:
                    pos = cfgfile.tell()
                    fieldvalue, newpos = get_struct(cfgfile, pos, nel, fname)
                    cfgfile.seek(newpos)
                else:
                    pos = cfgfile.tell()
                    fieldvalue, newpos = get_array(cfgfile, pos, nel, valtype)
                    cfgfile.seek(newpos)
            else:
                fieldvalue = string_to_datatype(valtype, valuestr)

            cfg.update({fieldname: fieldvalue})
        else:
            fileend = 1

    cfgfile.close()
    return cfg


def get_num_elements(dtype, nelstr):
    """
    Checks if data type is an array or a structure.

    Parameters
    ----------
    dtype : str
        data type specifier

    nelstr : str
        number of elements

    Returns
    -------
    nel : int
        number of elements if type is *ARR or STRUCT. 0 otherwise

    isstruct : bool
        true if the type is STRUCT

    """

    uptype = dtype.upper()
    isstruct = False
    nel = 0
    narr = uptype.count('ARR')
    if uptype == 'STRUCT':
        isstruct = True
    if (narr > 0) or (isstruct is True):
        nel = int(nelstr[0])

    return nel, isstruct


def string_to_datatype(dtype, strval):
    """
    Converts a string containing a value into its Python value

    Parameters
    ----------
    dtype : str
        data type specifier

    strval : str
        string value

    Returns
    -------
    val : scalar
        value contained in the string

    """

    uptype = dtype.upper()

    if uptype == 'BYTE':
        return int(strval[0])
    elif uptype == 'BOOL':
        return  bool(strval[0])
    elif uptype == 'INT':
        return int(strval[0])
    elif uptype == 'LONG':
        return int(strval[0])
    elif uptype == 'HEX':
        return int(strval[0])
    elif uptype == 'EXP':
        return float(strval[0])
    elif uptype == 'FLOAT':
        return float(strval[0])
    elif uptype == 'DOUBLE':
        return float(strval[0])
    elif uptype == 'STRING':
        # Replace all environement variables
        return os.path.expandvars(str(strval[0]))
    else:
        raise Exception("ERROR: Unexpected data type "+uptype)


def get_array(cfgfile, pos, nel, valtype):
    """
    reads an array in a config file

    Parameters
    ----------
    cfgfile : file object
        config file

    pos : int
        position in file object

    nel : int
        number of elements of the ray

    valtype : str
        type of array

    Returns
    -------
    arr : array
        array values

    newpos : int
        new position in file object

    """

    arr_type = get_array_type(valtype)
    if arr_type == 'STRING':
        arr = []
    else:
        arr = init_array(nel, arr_type)

    newpos = pos
    for i in range(nel):
        pos = cfgfile.seek(newpos)
        line = cfgfile.readline()

        # remove leading and trailing whitespace
        line = line.strip()

        line = line.partition('#')[0]  # Remove comments
        line = line.strip()

        vals = line.split()

        value = string_to_datatype(arr_type, vals)
        if arr_type == 'STRING':
            arr.append(value)
        else:
            arr[i] = value
        newpos = cfgfile.tell()

    return arr, newpos


def get_struct(cfgfile, pos, nels, fname):
    """
    reads an struct in a config file

    Parameters
    ----------
    cfgfile : file object
        config file
    pos : int
        position in file object
    nel : int
        number of elements of the ray
    fname : str
        config file name

    Returns
    -------
    struct : dict
        dictionary of struct values

    newpos : int
        new position in file object

    """

    struct = dict()
    newpos = pos
    for i in range(nels):
        pos = cfgfile.seek(newpos)
        line = cfgfile.readline()

        # remove leading and trailing whitespace
        line = line.strip()

        line = line.partition('#')[0]  # Remove comments
        line = line.strip()

        vals = line.split()

        sfieldname = vals[0]
        nvals = len(vals)

        if nvals < 3:
            raise Exception(
                "FILE FORMAT ERROR: file: " + fname +
                ", struct variable: " + sfieldname +
                ": Wrong number of elements!")

        svaltype = vals[1]
        svaluestr = vals[2:nvals]
        nel, isstruct = get_num_elements(svaltype, svaluestr)

        if nel > 0:
            if isstruct:
                pos = cfgfile.tell()
                sfieldvalue, newpos = get_struct(cfgfile, pos, nel, fname)
                cfgfile.seek(newpos)
            else:
                pos = cfgfile.tell()
                sfieldvalue, newpos = get_array(cfgfile, pos, nel, svaltype)
                cfgfile.seek(newpos)
        else:
            newpos = cfgfile.tell()
            sfieldvalue = string_to_datatype(svaltype, svaluestr)

        struct.update({sfieldname: sfieldvalue})

    return struct, newpos


def get_array_type(dtype):
    """
    Determines Python array type from the config file array type

    Parameters
    ----------
    dtype : str
        config file data type

    Returns
    -------
    pytype : str
        Python array type

    """

    uptype = dtype.upper()

    if uptype == 'BYTARR':
        return 'BYTE'
    elif uptype == 'INTARR':
        return 'INT'
    elif uptype == 'LONARR':
        return 'LON'
    elif uptype == 'HEXARR':
        return 'HEX'
    elif uptype == 'EXPARR':
        return 'EXP'
    elif uptype == 'FLTARR':
        return 'FLOAT'
    elif uptype == 'DBLARR':
        return 'DOUBLE'
    elif uptype == 'STRARR':
        return 'STRING'
    else:
        raise Exception("ERROR: Unexpected data type "+uptype)


def init_array(nel, dtype):
    """
    Initializes a Python array

    Parameters
    ----------
    nel : int
        number of elements in the array
    dtype : str
        config file data type

    Returns
    -------
    pyarr : array
        Python array

    """

    uptype = dtype.upper()

    if uptype == 'BYTE':
        return np.empty(nel, dtype='byte')
    elif uptype == 'INT':
        return np.empty(nel, dtype='int32')
    elif uptype == 'LONG':
        return np.empty(nel, dtype='int64')
    elif uptype == 'HEX':
        return np.empty(nel, dtype='H')
    elif uptype == 'EXP':
        return np.empty(nel, dtype='float64')
    elif uptype == 'FLOAT':
        return np.empty(nel, dtype='float64')
    elif uptype == 'DOUBLE':
        return np.empty(nel, dtype='double')
    elif uptype == 'STRING':
        return []
    else:
        raise Exception("ERROR: Unexpected array data type "+uptype)



"""
pyrad.io.read_data_mxpol
========================
Functions for reading radar mxpol data files
.. autosummary::
    :toctree: generated/
    classes - MXPOL:
        pyrad_MXPOL
    classes - MCH:
        pyrad_MCH
    utilities - read:
        row_stack
        findTimes
        int2float_radar
        readMXPOLRadData
        readCHRadData
    utilities - config:
        load_myconfig
        get_mymetadata
        get_elevation_metadata
        generate_radar_table
        generate_polvar_metadata
        convert_polvar_name
"""

import time
import importlib.util
import re
import os
import datetime
from copy import deepcopy
import warnings

import numpy as np
import netCDF4
try:
    import h5py
    _H5PY_AVAILABLE = True
except ImportError:
    _H5PY_AVAILABLE = False

class MissingOptionalDependency(Exception):
    """ Exception raised when a optional dependency is needed but not found. """
    pass

import pyart

# -------------------------- classes - MXPOL ------------------------------ #


class pyrad_MXPOL(pyart.core.Radar):
    def __init__(self, filename, field_names=None, max_range=np.Inf,
                 min_range=10000, pyrad_names=True):
        # find information based on filename
        all_files = [filename]
        fname_basename = os.path.basename(filename)

        if 'PPI' in fname_basename:
            scan_type = 'ppi'
        elif 'RHI' in fname_basename:
            scan_type = 'rhi'

        strdate = re.findall(r"([0-9]{8}-[0-9]{6})", fname_basename)[0]
        date = datetime.datetime.strptime(strdate, '%Y%m%d-%H%M%S')

        # if field name is None, take all available fields

        if field_names is None:
            field_names = ['Zh', 'Zdr', 'Kdp', 'Phidp', 'Rhohv', 'ZhCorr',
                           'ZdrCorr', 'RVel', 'Rvel', 'Sw', 'SNRh', 'SNRv', 'Psidp']

        # convert fieldname if necessary
        varnames = []
        for fieldname in field_names:
            newname = convert_polvar_name('LTE', fieldname)
            varnames.append(newname)

        # get labels, units etc
        long_names = []
        standard_names = []
        units = []
        vmin = []
        vmax = []

        for varname in varnames:
            metadata = generate_polvar_metadata(varname)
            standard_names.append(metadata['standard_name'])
            long_names.append(metadata['long_name'])
            units.append(metadata['units'])
            vmin.append(metadata['valid_min'])
            vmax.append(metadata['valid_max'])

        # initiate empty vectors
        N_sweeps = len(all_files)
        fields = {}
        fixed_angle = {}
        fixed_angle['data'] = np.zeros(N_sweeps, )

        sweep_start_ray_index = {}
        sweep_start_ray_index['data'] = []
        sweep_stop_ray_index = {}
        sweep_stop_ray_index['data'] = []

        for i, k in enumerate(varnames):
            fields[k] = {}
            fields[k]['data'] = []
            fields[k]['long_name'] = long_names[i]
            fields[k]['standard_name'] = standard_names[i]
            fields[k]['units'] = units[i]
            fields[k]['valid_min'] = vmin[i]
            fields[k]['valid_max'] = vmax[i]

        idx_start = 0
        idx_stop = 0
        elevations = []
        azimuths = []
        ranges = []
        nyquist = []

        # read data and create dictionaries
        for i in range(N_sweeps):
            metadata, data = readMXPOLRadData(
                all_files[i], varnames, max_range)
            if scan_type == 'rhi':
                fixed_angle['data'][i] = np.round(np.mean(data['azimuth']))
            elif scan_type == 'ppi':
                fixed_angle['data'][i] = np.round(np.mean(data['elevation']))

            [N_az, N_ranges] = data[varnames[0]].shape
            idx_stop = idx_start + N_az - 1
            sweep_start_ray_index['data'].append(idx_start)
            sweep_stop_ray_index['data'].append(idx_stop)
            idx_start = idx_stop + 1
            elevations.extend(list(data['elevation']))
            nyquist.extend([data['nyquist_vel']]*N_az)
            azimuths.extend(list(data['azimuth']))
            ranges.extend(list(data['range']))

            for j, v in enumerate(varnames):
                if v in data.keys():
                    if not(len(fields[v]['data'])):
                        fields[v]['data'] = data[v]
                    else:
                        fields[v]['data'] = row_stack(
                            fields[v]['data'], data[v])
                else:
                    print('Variable '+v+' was not found in file!')

        # mask NaNs

        for v in varnames:
            if not len(fields[v]['data']):
                # Remove variable
                fields.pop(v)
            else:
                fields[v]['data'] = np.ma.masked_equal(
                    fields[v]['data'], -99900.0)

        [a, N_ranges] = fields[varnames[0]]['data'].shape

        # create dictionaries according to pyART standard
        latitude = {'data': np.asarray([data['latitude']]),
                    'units': data['lat_units']}
        longitude = {'data': np.asarray([data['longitude']]),
                     'units': data['lon_units']}
        altitude = {'data': np.asarray([data['altitude']]),
                    'units': data['alt_units']}
        sweep_number = {'data': np.arange(0, len(all_files))}
        sweep_mode = {'data': np.asarray([scan_type]*N_sweeps)}
        instrument_parameters = {
            'nyquist_velocity': {'data': np.asarray(nyquist)}}
        azimuth = {'data': np.asarray(azimuths), 'units': data['azim_units']}
        rrange = {'data': np.asarray(ranges),
                  'units': data['range_units']}
        elevation = {'data': np.asarray(elevations),
                     'units': data['elev_units']}
        sweep_start_ray_index['data'] = np.asarray(
            sweep_start_ray_index['data'])
        sweep_stop_ray_index['data'] = np.asarray(
            sweep_stop_ray_index['data'])

        time_units = 'seconds since ' + str(date)
        time_data = {'data': data['time'], 'units': time_units}

        # change keys to match pyART metranet keys
        if pyrad_names:
            fields_copy = deepcopy(fields)
            for keys in fields_copy:
                newkey = fields[keys]['standard_name']
                fields[newkey] = fields.pop(keys)

        # Create PyART instance
        pyart.core.Radar.__init__(
            self, time_data, rrange, fields, metadata, scan_type, latitude,
            longitude, altitude, sweep_number, sweep_mode, fixed_angle,
            sweep_start_ray_index, sweep_stop_ray_index, azimuth, elevation,
            instrument_parameters=instrument_parameters)


# -------------------------- classes - IDL --------------------------- #

class pyrad_IDL(pyart.core.Radar):
    def __init__(self, filename, field_names=None, max_range=np.Inf,
                 min_range=10000):

        # find information based on filename
        all_files = [filename]
        fname_basename = os.path.basename(filename)

        fname = netCDF4.Dataset(filename)

        if 'PPI' in fname_basename:
            scan_type = 'ppi'
        elif 'RHI' in fname_basename:
            scan_type = 'rhi'
        strdate = re.findall(r"([0-9]{8}-[0-9]{6})", fname_basename)[0]
        date = datetime.datetime.strptime(strdate, '%Y%m%d-%H%M%S')

        # if field name is None, take all available fields

        if field_names is None:
            field_names = list(fname.variables.keys())

        # convert fieldname if necessary
        varnames = []
        for fieldname in field_names:
            newname = convert_polvar_name('IDL', fieldname)
            varnames.append(newname)

        # get labels, units etc
        long_names = []
        standard_names = []
        units = []
        vmin = []
        vmax = []

        for varname in varnames:
            metadata = generate_polvar_metadata(varname)
            standard_names.append(metadata['standard_name'])
            long_names.append(metadata['long_name'])
            units.append(metadata['units'])
            vmin.append(metadata['valid_min'])
            vmax.append(metadata['valid_max'])

        # initiate empty vectors
        N_sweeps = len(all_files)
        fields = {}
        fixed_angle = {}
        fixed_angle['data'] = np.zeros(N_sweeps, )

        sweep_start_ray_index = {}
        sweep_start_ray_index['data'] = []
        sweep_stop_ray_index = {}
        sweep_stop_ray_index['data'] = []

        for i, k in enumerate(varnames):
            fields[k] = {}
            fields[k]['data'] = []
            fields[k]['long_name'] = long_names[i]
            fields[k]['standard_name'] = standard_names[i]
            fields[k]['units'] = units[i]
            fields[k]['valid_min'] = vmin[i]
            fields[k]['valid_max'] = vmax[i]

        idx_start = 0
        idx_stop = 0
        elevations = []
        azimuths = []
        nyquist = []

        # read data and create dictionaries
        for i in range(N_sweeps):
            metadata, data = readIDLRadData(
                all_files[i], varnames, max_range)
            if scan_type == 'rhi':
                fixed_angle['data'][i] = np.round(np.mean(data['azimuth']))
            elif scan_type == 'ppi':
                fixed_angle['data'][i] = data['elevation'][0]

            [N_az, N_ranges] = data[varnames[0]].shape
            idx_stop = idx_start + N_az - 1
            sweep_start_ray_index['data'].append(idx_start)
            sweep_stop_ray_index['data'].append(idx_stop)
            idx_start = idx_stop + 1
            elevations.extend([data['elevation'][0]]*N_az)
            nyquist.extend([data['nyquist_vel']]*N_az)
            azimuths.extend(list(data['azimuth']))
            warnings.warn("Warning, sweep rank could not be found, using first rank")

            starttime, endtime = findTimes(1)
            interval = ((endtime-starttime)/N_az)
            #time_lapse = np.arange(starttime+(0.5*interval), endtime, interval)
            # because this is a single sweep
            time_lapse = np.around(
                np.arange(0.+(0.5*interval), endtime-starttime, interval))

            for j, v in enumerate(varnames):
                if v in data.keys():
                    if fields[v]['data'].size == 0:
                        fields[v]['data'] = data[v]
                    else:
                        fields[v]['data'] = row_stack(
                            fields[v]['data'], data[v])
                else:
                    print('Variable '+v+' was not found in file!')

        # mask NaNs

        for v in varnames:
            fields[v]['data'] = np.ma.masked_equal(
                fields[v]['data'], -99900.0)

        [a, N_ranges] = fields[varnames[0]]['data'].shape

        # create dictionaries according to pyART standard
        latitude = {'data': np.asarray([data['latitude']]),
                    'units': data['lat_units']}
        longitude = {'data': np.asarray([data['longitude']]),
                     'units': data['lon_units']}
        altitude = {'data': np.asarray([data['altitude']]),
                    'units': data['alt_units']}
        sweep_number = {'data': np.arange(0, len(all_files))}
        sweep_mode = {'data': np.asarray([scan_type]*N_sweeps)}
        instrument_parameters = {
            'nyquist_velocity': {'data': np.asarray(nyquist)}}
        azimuth = {'data': np.asarray(azimuths), 'units': data['azim_units']}
        rrange = {'data': np.arange(N_ranges)*data['resolution'],
                  'units': data['range_units']}
        elevation = {'data': np.asarray(elevations),
                     'units': data['elev_units']}
        sweep_start_ray_index['data'] = np.asarray(
            sweep_start_ray_index['data'])
        sweep_stop_ray_index['data'] = np.asarray(
            sweep_stop_ray_index['data'])

        time_units = 'seconds since ' + str(date)
        time_lapse = np.asarray(time_lapse)
        time_data = {'data': time_lapse, 'units': time_units}

        # change keys to match pyART metranet keys
        fields_copy = deepcopy(fields)
        for keys in fields_copy:
            newkey = fields[keys]['standard_name']
            fields[newkey] = fields.pop(keys)

        # Create PyART instance
        pyart.core.Radar.__init__(
            self, time_data, rrange, fields, metadata, scan_type, latitude,
            longitude, altitude, sweep_number, sweep_mode, fixed_angle,
            sweep_start_ray_index, sweep_stop_ray_index, azimuth, elevation,
            instrument_parameters=instrument_parameters)

# -------------------------- classes - MCH --------------------------- #

class pyrad_MCH(pyart.core.Radar):
    def __init__(self, filename, field_names=None, max_range=np.Inf):

        # find information based on filename
        all_files = [filename]
        N_sweeps = len(all_files)

        fname_basename = os.path.basename(filename)

        # Get name of radar
        index_letter = fname_basename[2]

        radar_info = generate_radar_table(index_letter)
        radar_name = radar_info['radarID']

        # Get radar resolution
        if fname_basename[1] == 'L':
            rres = 500.
        else:
            rres = 83.3

        scan_type = 'ppi'

        scandate = datetime.datetime.strptime(
            fname_basename[3:12], '%y%j%H%M')
        self.scan_date = scandate.timetuple()

        # if field name is None, take all available fields
        if field_names is None:
            field_names = ['Z', 'ZDR', 'ZV', 'V', 'W', 'RHO', 'CLUT', 'PHIDP']

        # convert fieldname if necessary
        varnames = []
        for fieldname in field_names:
            newname = convert_polvar_name('MCH', fieldname)
            varnames.append(newname)

        # get labels, units etc
        long_names = []
        standard_names = []
        units = []
        vmin = []
        vmax = []

        for varname in varnames:
            metadata = generate_polvar_metadata(varname)
            standard_names.append(metadata['standard_name'])
            long_names.append(metadata['long_name'])
            units.append(metadata['units'])
            vmin.append(metadata['valid_min'])
            vmax.append(metadata['valid_max'])

        # initiate empty vectors
        fields = {}
        fixed_angle = {}
        fixed_angle['data'] = np.zeros(N_sweeps, )

        sweep_start_ray_index = {}
        sweep_start_ray_index['data'] = []
        sweep_stop_ray_index = {}
        sweep_stop_ray_index['data'] = []

        for i, k in enumerate(varnames):
            fields[k] = {}
            fields[k]['data'] = []
            fields[k]['long_name'] = long_names[i]
            fields[k]['standard_name'] = standard_names[i]
            fields[k]['units'] = units[i]
            fields[k]['valid_min'] = vmin[i]
            fields[k]['valid_max'] = vmax[i]

        # Initialize
        idx_start = 0
        idx_stop = 0
        elevations = []
        azimuths = []
        nyquist = []
        time_lapse = []

        # read and organise data
        for i in range(N_sweeps):
            data = readCHRadData(
                all_files[i], radar_name, varnames, rres, max_range)
            fixed_angle['data'][i] = data['elevation']
            [N_ranges, N_az] = data[varnames[0]].shape
            idx_stop = idx_start + N_az - 1
            sweep_start_ray_index['data'].append(idx_start)
            sweep_stop_ray_index['data'].append(idx_stop)
            idx_start = idx_stop + 1
            elevations.extend([data['elevation']]*N_az)
            nyquist.extend([data['nyquist_vel']]*N_az)
            azimuths.extend(list(data['azimuth']))
            # create list of times at the center of each ray
            sweep_rank = 1
            print()
            starttime, endtime = findTimes(sweep_rank)
            interval = ((endtime-starttime)/len(list(data['azimuth'])))
            time_lapse.extend(np.arange(
                starttime+(0.5*interval), endtime, interval))
            for j, v in enumerate(varnames):
                if fields[v]['data'].size == 0:
                    fields[v]['data'] = data[v].T
                else:
                    fields[v]['data'] = row_stack(
                        fields[v]['data'], data[v].T)

        # mask nans
        for v in varnames:
            fields[v]['data'] = np.ma.array(
                fields[v]['data'], mask=np.isnan(fields[v]['data']))

        sweep_start_ray_index['data'] = np.asarray(
            sweep_start_ray_index['data'])
        sweep_stop_ray_index['data'] = np.asarray(
            sweep_stop_ray_index['data'])
        metadata = {}

        [a, N_ranges] = fields[varnames[0]]['data'].shape

        latitude = {'data': np.array([radar_info['coordinates'][0]]),
                    'units': "DegreesNorth"}
        longitude = {'data': np.array([radar_info['coordinates'][1]]),
                     'units': "DegreesEast"}
        altitude = {'data': np.array([radar_info['altitude']]),
                    'units': "MetersAboveSeaLevel"}
        sweep_number = {'data': np.arange(0, len(all_files))}
        sweep_mode = {'data': np.asarray(['ppi']*N_sweeps)}
        instrument_parameters = {
            'nyquist_velocity': {'data': np.array(nyquist)}}

        metadata['Source'] = (
            "Operational radar data processed at MeteoSwiss Locarno-Monti")
        metadata['Institution'] = (
            "MeteoSwiss, MDR, Locarno-Monti, Switzerland")
        metadata['History'] = [
            "created: %s, " % time.ctime(os.path.getctime(filename)) +
            "last modified: %s" % time.ctime(os.path.getmtime(filename))]
        metadata['ContactInformation'] = "marc.schneebeli@meteosvizzera.ch"

        azimuth = {'data': np.array(azimuths), 'units': "Degrees"}
        rrange = {'data': np.arange(N_ranges)*data['resolution'],
                  'units': "Meters"}
        elevation = {'data': np.array(elevations), 'units': "Degrees"}

        time_units = 'seconds since '+str(scandate)
        time_lapse = np.asarray(time_lapse)
        scantime = {'data': time_lapse, 'units': time_units}

        # change keys to match pyART metranet keys
        fields_copy = deepcopy(fields)
        for keys in fields_copy:
            newkey = fields[keys]['standard_name']
            fields[newkey] = fields.pop(keys)

        # Create PyART instance
        pyart.core.Radar.__init__(
            self, scantime, rrange, fields, metadata, scan_type, latitude,
            longitude, altitude, sweep_number, sweep_mode, fixed_angle,
            sweep_start_ray_index, sweep_stop_ray_index, azimuth, elevation,
            instrument_parameters=instrument_parameters)


# ----------------------- utilities - read --------------------- #

def row_stack(a1, a2):
    """
    Stacks data from subsequent sweeps, while padding "empty" columns from
    subsequent sweeps.
    Inputs
    ------
    a1: np.array
        destination array
    a2: np.array
        array which is added onto the first array
    Returns
    -------
    out: np.array
        stacked destination and additional array, with uniform shape
    """
    [N1, M1] = a1.shape
    [N2, M2] = a2.shape

    if M1 > M2:
        a2 = np.pad(a2, ((0, 0), (0, M1-M2)), mode='constant',
                    constant_values=-9999999)
    elif M2 < M1:
        a1 = np.pad(a2, ((0, 0), (0, M2-M1)), mode='constant',
                    constant_values=-9999999)

    out = np.vstack((a1, a2))
    out[out == -9999999] = np.nan

    return out


def findTimes(num_sweep):
    """
    Finds the times at the beginning and at the end of each sweep. Information
    comes from the elapsed time since the beginning of the volume scan, from
    the Rad4Alp: Specifications/ Request for Proposal (RFP) document.
    Inputs
    ------
    num_sweep: int
        rank of the sweep
    Returns
    -------
    elapsed_times[num_sweep][0]: float
        the elapsed time since the beginning of the volume scan at the
        beginning of the sweep
    elapsed_times[num_sweep][1]: float
        the elapsed time since the beginning of the volume scan at the end of
        the sweep
    """

    elapsed_times = {9: [0, 11.4],
                     7: [11.4, 22.8],
                     5: [22.8, 39.2],
                     3: [39.3, 60.5],
                     1: [60.5, 84.7],
                     19: [84.7, 97.2],
                     17: [97.2, 109.6],
                     15: [109.6, 121.6],
                     13: [121.6, 133.1],
                     11: [133.1, 144.4],
                     10: [144.4, 155.8],
                     8: [155.8, 172.2],
                     6: [172.2, 188.6],
                     4: [188.6, 204.9],
                     2: [204.9, 229.4],
                     20: [229.4, 241.9],
                     18: [241.9, 254.4],
                     16: [254.4, 266.6],
                     14: [266.6, 278.3],
                     12: [278.3, 289.9]}

    return elapsed_times[num_sweep][0], elapsed_times[num_sweep][1]


def int2float_radar(data, varname, index_angle):
    """
    Converts radar moments from bit to float
    Inputs
    ------
    data: np.array
        moment data as loaded from h5 file
    varname: str
        name of the moment (i.e. 'ZH')
    index_angle: int
        rank of the sweep-1 (converted to base 0)
    Returns
    -------
    output: np.array
        moment data converted to float
    """
    varname = convert_polvar_name('metranet', varname)
    NYQUIST_VEL = get_mymetadata('nyq_vel')

    output = np.zeros(data.shape)
    if varname in ['ZH', 'ZV', 'Z', 'ZHC']:
        output[data != 0] = (data[data != 0]-64)*0.5
        output[data == 0] = float('nan')
    elif varname == 'VEL':
        output[data != 0] = (data[data != 0]-128)/127*NYQUIST_VEL[index_angle]
        output[data == 0] = float('nan')
    elif varname == 'WID':
        output = data/255*NYQUIST_VEL[index_angle]
    elif varname in ['ZDR', 'ZDRC']:
        output[data != 0] = data[data != 0]*1.0/16.1259842 - 7.9375
        output[data == 0] = float('nan')
    elif varname == 'RHO':
        output[data != 0] = 1.003-10**(-(data[data != 0]-1.0)/100)
        output[data == 0] = float('nan')
    elif varname == 'PHI':
        output[data != 0] = (data[data != 0]-32768)/32767*180
        output[data == 0] = float('nan')
    elif varname == 'CLUT':
        output = data
    else:
        output = data
        warnings.warn(
            ("Warning, %s was not found and could not be converted")
            % (varname))

    return output


def readMXPOLRadData(filename, variableList, max_range=np.Inf, min_range=0):
    """
    Reads a netcdf containing processed radar data in polar coordinates
    Parameters
    ----------
    filename: str
        complete path of the file
    variableList: list
        list of variables to be read
    Returns
    -------
    varPol: dict
        dictionary containing the variables, the azimuth and the range
    metadata: dict
        dictionary containing the metadata of the file
    """

    varPol = {}
    metadata = {}
    ncid = netCDF4.Dataset(filename)

    time_data = ncid.variables['Time']
    time_data -= time_data[0]  # To get time in seconds from beginning of scan

    rrange = ncid.variables['Range'][:]

    # Get indexes between min_range and max_range
    idx2keep = np.where(np.logical_and(
        rrange < max_range, rrange > min_range))[0]
    rrange = rrange[idx2keep]

    # Get variables in polar coordinates
    for varname in variableList:
        try:
            varPol[varname] = ncid.variables[varname][:].T
        except:
            pass

    varPol['resolution'] = ncid.__dict__['RangeResolution-value']
    varPol['range'] = rrange
    varPol['range_units'] = ncid.__dict__['RangeResolution-unit']
    varPol['azimuth'] = ncid.variables['Azimuth'][:]
    try:
        varPol['azim_units'] = ncid.__dict__['Azimuth-unit']
    except KeyError:
        varPol['azim_units'] = ncid.variables['Azimuth'].Units
    varPol['elevation'] = ncid.variables['Elevation'][:]
    try:
        varPol['elev_units'] = ncid.__dict__['Elevation-unit']
    except KeyError:
        varPol['elev_units'] = ncid.variables['Elevation'].Units
    varPol['nyquist_vel'] = ncid.__dict__['NyquistVelocity-value']
    varPol['longitude'] = ncid.__dict__['Longitude-value']
    varPol['lon_units'] = ncid.__dict__['Longitude-unit']
    varPol['latitude'] = ncid.__dict__['Latitude-value']
    varPol['lat_units'] = ncid.__dict__['Latitude-unit']
    varPol['altitude'] = ncid.__dict__['Altitude-value']
    varPol['alt_units'] = ncid.__dict__['Altitude-unit']
    varPol['time'] = time_data

    metadata['Source'] = ncid.__dict__['Source']
    metadata['Institution'] = ncid.__dict__['Institution']
    metadata['History'] = ncid.__dict__['History']
    metadata['ContactInformation'] = ncid.__dict__['ContactInformation']

    # Close netcdf
    ncid.close()

    return metadata, varPol

def readIDLRadData(filename, variableList, max_range=np.Inf, min_range=0):
    """
    Reads a netcdf containing IDL processed radar data in polar coordinates
    Parameters
    ----------
    filename: str
        complete path of the file
    variableList: list
        list of variables to be read
    Returns
    -------
    varPol: dict
        dictionary containing the variables, the azimuth and the range
    metadata: dict
        dictionary containing the metadata of the file
    """

    varPol = {}
    metadata = {}
    ncid = netCDF4.Dataset(filename)

    time_data = ncid.variables['Time']
    time_data -= time_data[0]  # To get time in seconds from beginning of scan

    rrange = ncid.variables['Range'][:]

    # Get indexes between min_range and max_range
    idx2keep = np.where(np.logical_and(
        rrange < max_range, rrange > min_range))[0]
    rrange = rrange[idx2keep]

    # Get variables in polar coordinates
    for varname in variableList:
        try:
            varPol[varname] = ncid.variables[varname][:].T
        except:
            pass

    varPol['resolution'] = ncid.__dict__['RangeResolution-value']
    varPol['range'] = rrange
    varPol['range_units'] = ncid.__dict__['RangeResolution-unit']
    # because this data seems to be on -180 to 180
    #varPol['azimuth'] = (ncid.variables['Azimuth'][:] + 180)%360
    varPol['azimuth'] = ncid.variables['Azimuth'][:]
    try:
        varPol['azim_units'] = ncid.__dict__['Azimuth-unit']
    except KeyError:
        varPol['azim_units'] = ncid.variables['Azimuth'].Units
    varPol['elevation'] = ncid.variables['Elevation'][:]
    try:
        varPol['elev_units'] = ncid.__dict__['Elevation-unit']
    except KeyError:
        varPol['elev_units'] = ncid.variables['Elevation'].Units
    varPol['nyquist_vel'] = ncid.__dict__['NyquistVelocity-value']
    varPol['longitude'] = ncid.__dict__['Longitude-value']
    varPol['lon_units'] = ncid.__dict__['Longitude-unit']
    varPol['latitude'] = ncid.__dict__['Latitude-value']
    varPol['lat_units'] = ncid.__dict__['Latitude-unit']
    varPol['altitude'] = ncid.__dict__['Altitude-value']
    varPol['alt_units'] = ncid.__dict__['Altitude-unit']
    varPol['time'] = time_data

    metadata['Source'] = ncid.__dict__['Source']
    metadata['Institution'] = ncid.__dict__['Institution']
    metadata['History'] = ncid.__dict__['History']
    metadata['ContactInformation'] = ncid.__dict__['ContactInformation']

    # Close netcdf
    ncid.close()

    return metadata, varPol


def readCHRadData(filename, radar_name, variableList, radial_resolution,
                  max_range=np.Inf, min_range=0):
    """
    Reads a HDF5 file containing processed radar data in polar coordinates
    Parameters
    ----------
    filename: str
        complete path of the file
    radar_name: str
        name of MCH radar
    variableList: list
        list of variables to be read
    radial_resolution: float
        resolution of the radar in metres (i.e. high: 83.3, low: 500.)
    max_range: float
        maximum range upto which to read data
    min_range: float
        mimimum range from which to read data
    Returns
    -------
    varPol: dict
        the projected variables, the azimuth and the range
    """
    # check that h5py library is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use readCHRadData but is not installed")

    varPol = {}
    h5id = h5py.File(filename, 'r')

    ELEVATION_ANGLES = get_elevation_metadata(radar_name)
    radar_info = generate_radar_table(radar_name)
    ANG_RES = radar_info['dbbeam']
    NYQUIST_VEL = get_mymetadata('nyq_vel')

    # Get dimensions
    siz = h5id['moments']['Z'].shape
    rng = np.arange(0, siz[1])*radial_resolution
    idx2keep = np.where(np.logical_and(
        rng < max_range, rng > min_range))[0]
    rng = rng[idx2keep]
    azimuth = np.arange(0, siz[0])*ANG_RES
    index_angle = int(re.findall(r"\.([0-9]{3})\.", filename)[0])-1
    elevation = ELEVATION_ANGLES[index_angle]
    # Get variables in polar coordinates
    for varname in variableList:
        varname = convert_polvar_name('MCH', varname)
        data = []
        data = h5id['moments'][varname][:].T
        data = np.asarray(data)
        data = data.astype(float)
        clut = h5id['moments']['CLUT'][:].T
        data[clut >= 100] = float('nan')  # Remove clutter
        data = data[idx2keep, :]
        varPol[varname] = int2float_radar(data, varname, index_angle)

    varPol['resolution'] = rng[3]-rng[2]
    varPol['range'] = rng
    varPol['azimuth'] = azimuth
    varPol['elevation'] = elevation
    varPol['nyquist_vel'] = NYQUIST_VEL[index_angle]
    # Close netcdf
    h5id.close()

    return varPol


# ------------------------ utilities - config ------------------------- #

_dirname = os.path.dirname(__file__)
_DEFAULT_CONFIG_FILE = os.path.join(_dirname, 'mxpol_config.py')


def load_myconfig(filename=None):
    """
    Load configuration from a config file.
    Parameters
    ----------
    filename: str
        Filename of the configuration file. If None the default configuration
        file is loaded from the directory.
    Returns
    -------
    _DEFAULT_METADATA: dict
        Dictionary with metadata
    """

    if filename is None:
        filename = _DEFAULT_CONFIG_FILE

    # private:

    global cfile
    global _DEFAULT_POLARNAMES
    global _DEFAULT_METADATA
    global _DEFAULT_RADAR_INFO

    spec = importlib.util.spec_from_file_location("metadata_config", 
                                                  filename)
    cfile = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(cfile)
    
    _DEFAULT_METADATA = cfile.MY_METADATA
    _DEFAULT_POLARNAMES = cfile.MY_POLARNAMES
    _DEFAULT_RADAR_INFO = cfile.RADAR_INFO

    return _DEFAULT_METADATA


def get_mymetadata(p, filename=None):
    """
    Return a dictionary of metadata for a given parameter, p.
    An empty dictionary will be returned if no metadata dictionary exists for
    parameter p.
    Parameters
    ----------
    p: str
        parameter name (i.e. Polvar) for which to return metadata
    filename: str
        Filename of the configuration file. If None the default configuration
        file is loaded from the directory.
    Returns
    -------
    _DEFAULT_METADATA[p].copy(): dict
        a copy of the parameter of interest from the metadata dictionary
    """
    load_myconfig(filename=filename)

    if p in _DEFAULT_METADATA:
        return _DEFAULT_METADATA[p].copy()

    return {}


def get_elevation_metadata(radarname, filename=None):
    """
    Gets the elevation angles for each sweep from the configuration file
    Inputs
    ------
    radarname: str
        name of the radar for which to retrieve elevation angles
    filename: str
        name of the configuration file, if None, the default configuration
        file is used
    Returns
    -------
    _DEFAULT_RADAR_INFO['elevations'][radarname]: list
        list of elevation angles in degrees
    or None if not available
    """
    load_myconfig(filename=filename)

    if radarname in _DEFAULT_RADAR_INFO['elevations']:
        return _DEFAULT_RADAR_INFO['elevations'][radarname]
    else:
        print(("no elevation angles in configfile for radar %s") % (radarname))


def generate_radar_table(radarname, filename=None):
    """
    Generates a table with basic radar info, based on the given (or default)
    configfile
    Parameters
    ----------
    radarname: str
        name of the radar (i.e. 'ALB' or 'A', 'MXPOL' etc)
    filename: str
        path and name of the configfile, if None, the default configfile is
        used
    Returns
    -------
    radar_table: dict
        table containing basic radar info
    """
    load_myconfig(filename=filename)

    if radarname in _DEFAULT_RADAR_INFO['radarID']:
        radarname = _DEFAULT_RADAR_INFO['radarID'][radarname]
        radar_table = get_mymetadata('Radar_info', filename=filename)
        for key in radar_table:
            if key in _DEFAULT_RADAR_INFO:
                radar_table[key] = _DEFAULT_RADAR_INFO[key][radarname]
            else:
                radar_table[key] = None
        return radar_table

    return None


def generate_polvar_metadata(polvar, filename=None):
    """
    Generates a dictionary with metadata for a polarimetric variable
    Parameters
    ----------
    polvar: str
        polatimetric variable of interest
    filename: str
        Filename of the configuration file. If None the default configuration
        file is loaded from the directory.
    Returns
    -------
    polvar_metadata: dict
        dictionary with metatdata for polarimetric variable of interest
    """
    load_myconfig(filename=filename)
    polvar = convert_polvar_name('LTE', polvar)

    if polvar in _DEFAULT_POLARNAMES:
        (standard_name, long_name, units, valid_min, valid_max,
         plot_interval) = _DEFAULT_POLARNAMES[polvar]
    else:
        (standard_name, long_name, units, valid_min, valid_max,
         plot_interval) = None, None, None, None, None, None

    polvar_metadata = get_mymetadata('Polvar', filename)
    polvar_metadata['units'] = units
    polvar_metadata['standard_name'] = standard_name
    polvar_metadata['short_name'] = convert_polvar_name('MCH', polvar)
    polvar_metadata['long_name'] = long_name
    polvar_metadata['valid_min'] = valid_min
    polvar_metadata['valid_max'] = valid_max
    polvar_metadata['plot_interval'] = plot_interval

    return polvar_metadata


def convert_polvar_name(convention, polvar):
    """
    Finds the correct variable name for a given convention (MXPOL, MCH) and
    a given variable name which was spelled with a different case or
    according to a different convention. For example, MXPOL convention uses
    'Z' for the reflectivity variable, but if a user inserted 'Zh' this
    function will convert it to 'Z'.
    Parameters
    ----------
    convention : str, destination convention; either MCH or LTE
    polvar : str, key of polarimetric variable to be converted
    Returns
    -------
    mykey : str, polarimertric variable key as used within the ProfileLab
        toolbox context
    """
    # Generate dictionary for the conversion
    metranet_list = [
        'ZH', 'ZV', 'ZDR', 'PHI', 'VEL', 'VEL', 'WID', 'RHO', 'CLUT', 'MPH',
        'STA1', 'STA2', 'WBN', 'ZHC', 'ZDRC', 'ZDRP', 'Kdpc', 'Rhohvc']
    MCH_list = [
        'Z', 'ZV', 'ZDR', 'PHIDP', 'V', 'V', 'W', 'RHO', 'CLUT', 'MPH', 'STA1',
        'STA2', 'WBN', 'Zhc', 'Zdrc', 'Hydrometeor_type_from_Besic1', 'Kdpc', 'RHOC']
    # ZhCorr and ZdrCorr have been changed to Zhc and Zdrc!
    LTE_list = [
        'Zh', 'Zv', 'Zdr', 'Phidp', 'RVel', 'Rvel', 'Sw', 'Rhohv', 'Clut', 'mph',
        'sta1', 'sta2', 'wbn', 'Zhc', 'Zdrc', 'Hydroclass', 'Kdpc', 'Rhohvc']
    IDL_list = [
        'Zh', 'Zv', 'Zdr', 'Phidp_raw', 'V', 'V', 'W', 'uRhohv', 'CLUT', 'MPH', 'STA1',
        'STA2', 'WBN', 'Zhc', 'Zdrc', 'TYPECLUS2', 'Kdpc', 'Rhohvc']
    pyrad_list = [
        'reflectivity', 'reflectivity_vv', 'differential_reflectivity',
        'differential_phase', 'velocity', 'velocity', 'spectrum_width',
        'uncorrected_cross_correlation_ratio', 'radar_echo_id', 'MPH',
        'STA1', 'STA2', 'WBN', 'corrected_reflectivity',
        'corrected_differential_reflectivity', 'radar_echo_classification',
        'corrected_specific_differential_phase',
        'corrected_cross_correlation_ratio']

    convertkeys = {}
    convertkeys['MCH'] = {}
    convertkeys['LTE'] = {}
    convertkeys['metranet'] = {}
    convertkeys['IDL'] = {}
    convertkeys['pyrad'] = {}

    for i, MCH in enumerate(MCH_list):
        convertkeys['MCH'][MCH] = [
            LTE_list[i], metranet_list[i], IDL_list[i], pyrad_list[i]]

    convertkeys['LTE'] = {}
    for i, LTE in enumerate(LTE_list):
        convertkeys['LTE'][LTE] = [
            MCH_list[i], metranet_list[i], IDL_list[i], pyrad_list[i]]

    for i, metranet in enumerate(metranet_list):
        convertkeys['metranet'][metranet] = [
            MCH_list[i], LTE_list[i], IDL_list[i], pyrad_list[i]]

    for i, IDL in enumerate(IDL_list):
        convertkeys['IDL'][IDL] = [
            metranet_list[i], LTE_list[i], MCH_list[i], pyrad_list[i]]

    for i, pyrad in enumerate(pyrad_list):
        convertkeys['pyrad'][pyrad] = [
            metranet_list[i], LTE_list[i], MCH_list[i], IDL_list[i]]

    # translate between conventions
    mykey = polvar

    for key, value in convertkeys[convention].items():
        if polvar in value:
            mykey = key
            break

    return mykey
